from __future__ import annotations

import inspect
import os
import sys
from itertools import product
from typing import Any, Optional

from cerberus import Validator
from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig
from omegaconf.listconfig import ListConfig
from rich import print as rich_print
from torch.nn import Module
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler

from . import DDict  # NOQA
from .utils import get_optimizer, get_scheduler  # NOQA

sys.path.append(".")


__all__ = ["Configuration", "ConfigurationError", "OptimizerConfigurator"]


class ConfigurationError(Exception):
    """Exception raised when the configuration is not valid."""

    def __init__(self, message: str = "Invalid configuration.", errors_dict: Optional[dict] = None):
        super().__init__(message)
        rich_print(f"[red]ConfigurationError:[/red] {message}")
        if errors_dict is not None:
            rich_print(errors_dict)


class Configuration(DDict):
    """Fluke configuration class.
    This class is used to store the configuration of an experiment. The configuration must adhere to
    a specific structure. The configuration is validated when the class is instantiated.

    Args:
        config_exp_path (str): The path to the experiment configuration file.
        config_alg_path (str): The path to the algorithm configuration file.
        force_validation (bool, optional): Whether to force the validation of the configuration.
            Defaults to ``True``.

    Raises:
        ValueError: If the configuration is not valid.
    """

    def __init__(
        self,
        config_exp_path: str = None,
        config_alg_path: str = None,
        force_validation: bool = True,
    ):
        super().__init__()

        if config_exp_path is not None and os.path.exists(config_exp_path):
            cfg_exp = OmegaConf.load(config_exp_path)
            self.update(DDict(**cfg_exp))

        if config_alg_path is not None and os.path.exists(config_alg_path):
            cfg_alg = OmegaConf.load(config_alg_path)
            self.update(method=DDict(**cfg_alg))

        if force_validation:
            self._validate()

    @classmethod
    def from_dict(cls, cfg_dict: dict | DictConfig) -> Configuration:
        """Create a configuration from a dictionary.

        Args:
            cfg_dict (dict | DictConfig): The dictionary.

        Returns:
            Configuration: The configuration.
        """
        cfg = Configuration(force_validation=False)
        cfg.update(**cfg_dict)
        cfg._validate()
        return cfg

    @classmethod
    def fromkeys(cls, *args, **kwargs) -> Configuration:
        # Hides the fromkeys method of dict
        raise AttributeError("'Configuration' class has no method 'fromkeys'")

    def to_dict(self) -> dict:
        """Convert the configuration to a dictionary.

        Returns:
            dict: The dictionary.
        """

        def _to_dict(ddict: DDict) -> Any:
            if not isinstance(ddict, dict):
                if isinstance(ddict, type):
                    return ddict.__name__
                return ddict
            return {k: _to_dict(v) for k, v in ddict.items()}

        return _to_dict(self)

    @classmethod
    def sweep(cls, config_exp_path: str, config_alg_path: str) -> list[Configuration]:
        """Generate configurations from a sweep.
        This method is used to generate configurations from a sweep. The sweep is defined by the
        experiment configuration file. The method yields a configuration for each combination of
        hyperparameters.

        Args:
            config_exp_path (str): The path to the experiment configuration file.
            config_alg_path (str): The path to the algorithm configuration file.

        Returns:
            list[Configuration]: A list of configurations.
        """
        cfgs = Configuration(config_exp_path, config_alg_path, force_validation=False)
        all_configs = Configuration.__sweep(cfgs)
        return [Configuration.from_dict(cfg) for cfg in all_configs]

    @staticmethod
    def __sweep(cfgs: DDict | dict) -> list[Configuration]:
        """Generate configurations from a sweep.
        This method is used to generate configurations from a sweep. The sweep is defined by the
        experiment configuration file. The method yields a configuration for each combination of
        hyperparameters.

        Args:
            cfgs (DDict | dict): The configuration.

        Returns:
            list[Configuration]: A list of configurations.
        """
        normalized = {
            k: v if isinstance(v, (list, dict, ListConfig)) else [v] for k, v in cfgs.items()
        }

        for k, v in normalized.items():
            if isinstance(v, dict):
                nested_combos = Configuration.__sweep(v)
                normalized[k] = nested_combos

        keys = normalized.keys()
        values = normalized.values()

        all_combinations = []
        for combo in product(*values):
            combined = {k: v for k, v in zip(keys, combo)}
            all_combinations.append(combined)

        return all_combinations

    @property
    def client(self) -> DDict:
        """Get quick access to the client's hyperparameters.

        Returns:
            DDict: The client's hyperparameters.
        """
        return self.method.hyperparameters.client

    @property
    def server(self) -> DDict:
        """Get quick access to the server's hyperparameters.

        Returns:
            DDict: The server's hyperparameters.
        """
        return self.method.hyperparameters.server

    @property
    def model(self) -> DDict:
        """Get quick access to the model hyperparameters.

        Returns:
            DDict: The model hyperparameters.
        """
        return self.method.hyperparameters.model

    __SCHEMA = {
        "data": {
            "type": "dict",
            "schema": {
                "dataset": {
                    "type": "dict",
                    "required": True,
                    "schema": {
                        "name": {"type": "string", "required": True},
                        "path": {
                            "type": "string",
                            "required": False,
                            "default": "./data",
                        },
                    },
                },
                "distribution": {
                    "type": "dict",
                    "required": True,
                    "schema": {"name": {"type": "string", "required": True}},
                },
                "sampling_perc": {
                    "type": "float",
                    "required": False,
                    "min": 0.001,
                    "max": 1.0,
                    "default": 1.0,
                },
                "client_split": {
                    "type": "float",
                    "required": False,
                    "min": 0.0,
                    "max": 1.0,
                    "default": 0.0,
                },
                "keep_test": {"type": "boolean", "required": False, "default": True},
                "server_test": {"type": "boolean", "required": False, "default": True},
                "server_split": {
                    "type": "float",
                    "required": False,
                    "min": 0.0,
                    "max": 1.0,
                    "default": 0.0,
                },
                "uniform_test": {
                    "type": "boolean",
                    "required": False,
                    "default": False,
                },
            },
        },
        "exp": {
            "type": "dict",
            "schema": {
                "device": {
                    "required": False,
                    "default": "cpu",
                    "anyof": [
                        {
                            "type": "string",
                            "anyof": [
                                {"allowed": ["cpu", "cuda", "mps"]},
                                {"regex": "^cuda:[0-9]+$"},
                            ],
                        },
                        {"type": "list"},
                    ],
                },
                "seed": {"type": "integer", "required": True, "default": 42},
                "inmemory": {"type": "boolean", "required": True, "default": True},
            },
        },
        "eval": {
            "type": "dict",
            "schema": {
                "task": {
                    "type": "string",
                    "required": False,
                    "default": "classification",
                    "allowed": ["classification"],
                },
                "eval_every": {
                    "type": "integer",
                    "required": False,
                    "default": 1,
                    "min": 1,
                },
                "pre_fit": {"type": "boolean", "required": False, "default": False},
                "post_fit": {"type": "boolean", "required": False, "default": True},
                "locals": {"type": "boolean", "required": False, "default": False},
                "server": {"type": "boolean", "required": False, "default": True},
            },
        },
        "logger": {
            "type": "dict",
            "schema": {"name": {"type": "string", "required": False, "default": "Log"}},
        },
        "protocol": {
            "type": "dict",
            "schema": {
                "eligible_perc": {
                    "type": "float",
                    "required": True,
                    "min": 0.0,
                    "max": 1.0,
                },
                "n_clients": {"type": "integer", "required": True, "min": 1},
                "n_rounds": {"type": "integer", "required": True, "min": 1},
            },
        },
        "method": {
            "type": "dict",
            "schema": {
                "hyperparameters": {
                    "type": "dict",
                    "schema": {
                        "client": {
                            "type": "dict",
                            "schema": {
                                "batch_size": {
                                    "type": "integer",
                                    "required": True,
                                    "min": 0,
                                },
                                "local_epochs": {
                                    "type": "integer",
                                    "required": True,
                                    "min": 1,
                                },
                                "loss": {"type": "string", "required": True},
                                "optimizer": {
                                    "type": "dict",
                                    "schema": {
                                        "name": {
                                            "type": "string",
                                            "required": False,
                                            "default": "SGD",
                                        },
                                        "lr": {
                                            "type": "float",
                                            "required": False,
                                            "default": 0.01,
                                        },
                                    },
                                },
                                "scheduler": {
                                    "type": "dict",
                                    "schema": {
                                        "name": {
                                            "type": "string",
                                            "required": False,
                                            "default": "StepLR",
                                        }
                                    },
                                },
                            },
                        },
                        "server": {"type": "dict"},
                        "model": {"type": "string", "required": True},
                    },
                },
                "name": {"type": "string", "required": True},
            },
        },
    }

    @staticmethod
    def __repair_save(data: dict) -> tuple:
        if "save" not in data:
            return {}, []

        save_valid = Validator()
        save_valid.schema = {
            "save_every": {"type": "integer", "default": 1, "min": 1},
            "path": {"type": "string", "default": "./models"},
            "global_only": {"type": "boolean", "default": False},
        }
        save_valid.allow_unknown = False
        valid_result = save_valid.validate(data["save"])
        if not valid_result:
            return None, save_valid.errors

        return save_valid.document, []

    def _validate(self) -> None:

        validator = Validator()
        validator.schema = self.__SCHEMA
        validator.allow_unknown = True

        cfg_dict = self.to_dict()
        valid_result = validator.validate(cfg_dict)
        save_valid_result, save_errors = Configuration.__repair_save(cfg_dict)

        errors = validator.errors
        if save_errors:
            errors.update(save=save_errors)

        if not valid_result:
            raise ConfigurationError(errors_dict=errors)

        clean_cfg = validator.document
        clean_cfg["save"] = save_valid_result

        self.update(clean_cfg)

    def verbose(self) -> str:
        return super().__str__()


class OptimizerConfigurator:
    """This class is used to configure the optimizer and the learning rate scheduler.

    Attributes:
        optimizer (type[Optimizer]): The optimizer class.
        scheduler (type[LRScheduler]): The learning rate scheduler class.
        optimizer_cfg (DDict): The optimizer keyword arguments.
        scheduler_cfg (DDict): The scheduler keyword arguments.
    """

    def __init__(self, optimizer_cfg: DDict | dict, scheduler_cfg: DDict | dict = None):
        """Initialize the optimizer configurator. In both the ``optimizer_cfg`` and the
        ``scheduler_cfg`` dictionaries, the key "name" is used to specify the optimizer and the
        scheduler, respectively. If not present, the default optimizer is the ``SGD`` optimizer, and
        the default scheduler is the ``StepLR`` scheduler. The other keys are the keyword arguments
        for the optimizer and the scheduler, respectively.

        Note:
            In the key "name" of the ``optimizer_cfg`` dictionary, you can specify the optimizer
            class directly, instead of its string name. Same for the scheduler.

        Args:
            optimizer_cfg (dict, DDict): The optimizer class.
            scheduler_cfg (dict or DDict, optional): The scheduler keyword arguments. If not
                specified, the default scheduler is the ``StepLR`` scheduler with a step size of 1
                and a gamma of 1. Defaults to ``None``.
        """

        self.optimizer_cfg: DDict | None = None
        self.scheduler_cfg: DDict | None = None
        self.optimizer: type[Optimizer] | None = None
        self.scheduler: type[LRScheduler] | None = None

        if isinstance(optimizer_cfg, DDict):
            self.optimizer_cfg = optimizer_cfg
        elif isinstance(optimizer_cfg, dict):
            self.optimizer_cfg = DDict(**optimizer_cfg)
        else:
            raise ValueError("Invalid optimizer configuration.")

        if scheduler_cfg is None:
            self.scheduler_cfg = DDict(name="StepLR", step_size=1, gamma=1)
        elif isinstance(scheduler_cfg, DDict):
            self.scheduler_cfg = scheduler_cfg
        elif isinstance(scheduler_cfg, dict):
            self.scheduler_cfg = DDict(**scheduler_cfg)
        else:
            raise ValueError("Invalid scheduler configuration.")

        if isinstance(self.optimizer_cfg.name, str):
            self.optimizer = get_optimizer(self.optimizer_cfg.name)
        elif inspect.isclass(self.optimizer_cfg.name) and issubclass(
            self.optimizer_cfg.name, Optimizer
        ):
            self.optimizer = self.optimizer_cfg.name
        else:
            raise ValueError("Invalid optimizer name. Must be a string or an optimizer class.")

        if "name" not in self.scheduler_cfg:
            self.scheduler = get_scheduler("StepLR")
        elif isinstance(self.scheduler_cfg.name, str):
            self.scheduler = get_scheduler(self.scheduler_cfg.name)
        elif inspect.isclass(self.scheduler_cfg.name) and issubclass(
            self.scheduler_cfg.name, LRScheduler
        ):
            self.scheduler = self.scheduler_cfg.name
        else:
            raise ValueError("Invalid scheduler name. Must be a string or a scheduler class.")

        self.scheduler_cfg = self.scheduler_cfg.exclude("name")
        self.optimizer_cfg = self.optimizer_cfg.exclude("name")

    def __call__(self, model: Module, filter_fun: Optional[callable] = None, **override_kwargs):
        """Creates the optimizer and the scheduler.

        Args:
            model (Module): The model whose parameters will be optimized.
            filter_fun (callable): This must be a function of the model and it must returns the set
                of parameters that the optimizer will consider.
            override_kwargs (dict): The optimizer's keyword arguments to override the default ones.

        Returns:
            tuple[Optimizer, StepLR]: The optimizer and the scheduler.
        """
        if override_kwargs:
            self.optimizer_cfg.update(**override_kwargs)

        if filter_fun is None:
            optimizer = self.optimizer(
                filter(lambda p: p.requires_grad, model.parameters()),
                **self.optimizer_cfg,
            )
        else:
            optimizer = self.optimizer(filter_fun(model), **self.optimizer_cfg)
        scheduler = self.scheduler(optimizer, **self.scheduler_cfg)
        return optimizer, scheduler

    def __str__(self, indent: int = 0) -> str:
        strsched = self.scheduler.__name__
        indentstr = " " * (indent + 7)
        to_str = f"OptCfg({self.optimizer.__name__},\n{indentstr}"
        to_str += f",\n{indentstr}".join([f"{k}={v}" for k, v in self.optimizer_cfg.items()])
        to_str += f",\n{indentstr}{strsched}("
        indentstr = indentstr + " " * (len(strsched) + 1)
        to_str += f",\n{indentstr}".join([f"{k}={v}" for k, v in self.scheduler_cfg.items()])
        to_str += "))"
        return to_str

    def __repr__(self, indent: int = 0) -> str:
        return self.__str__(indent=indent)

    def __getstate__(self) -> dict:
        return self.__dict__

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)



"""Implementation of the [LG-FedAVG20]_ algorithm.

References:
    .. [LG-FedAVG20] Paul Pu Liang, Terrance Liu, Liu Ziyin, Nicholas B. Allen, Randy P. Auerbach,
       David Brent, Ruslan Salakhutdinov, Louis-Philippe Morency. Think Locally, Act Globally:
       Federated Learning with Local and Global Representations. In arXiv (2020).
       URL: https://arxiv.org/abs/2001.01523
"""

import sys
from typing import Sequence

# from torch.nn import CrossEntropyLoss
from torch.nn.modules import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import PersonalizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderHeadNet, HeadGlobalEncoderLocalNet  # NOQA
from ..server import Server  # NOQA
from ..utils import get_model  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA

__all__ = ["LGFedAVGClient", "LGFedAVGServer", "LGFedAVG"]

# The implementation is almost identical to FedPerClient
# The difference lies in the part of the network that is local and the part that is global.
# One is the inverse of the other.


class LGFedAVGClient(Client):

    def __init__(
        self,
        index: int,
        model: EncoderHeadNet,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: Module,  # In the paper it is fixed to CrossEntropyLoss
        local_epochs: int = 3,
        fine_tuning_epochs: int = 0,
        **kwargs,
    ):
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            **kwargs,
        )
        self.model = HeadGlobalEncoderLocalNet(model)
        self._save_to_cache()

    def send_model(self) -> None:
        self.channel.send(
            Message(self.model.get_global(), "model", self.index, inmemory=True),
            "server",
        )

    def receive_model(self) -> None:
        msg = self.channel.receive(self.index, "server", msg_type="model")
        safe_load_state_dict(self.model.get_global(), msg.payload.state_dict())


class LGFedAVGServer(Server):

    def __init__(
        self,
        model: Module,
        test_set: FastDataLoader,  # not used
        clients: Sequence[Client],
        weighted: bool = False,
    ):
        super().__init__(model=model, test_set=None, clients=clients, weighted=weighted)


class LGFedAVG(PersonalizedFL):

    def get_client_class(self) -> type[Client]:
        return LGFedAVGClient

    def get_server_class(self) -> type[Server]:
        return LGFedAVGServer



"""Implementation of the GEAR [GEAR22]_ algorithm.

References:
    .. [GEAR22] Chen Chen, Jie Zhang, Lingjuan Lyu. GEAR: A Margin-based Federated Adversarial
       Training Approach. In FL@AAAI (2022).
       URL: https://federated-learning.org/fl-aaai-2022/Papers/FL-AAAI-22_paper_34.pdf

"""

import sys
from typing import Literal

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import clear_cuda_cache  # NOQA
from . import CentralizedFL  # NOQA

__all__ = ["GEARClient", "GEAR", "MarginBasedCrossEntropyLoss"]


class MarginBasedCrossEntropyLoss(nn.Module):
    """Margin-based cross-entropy loss."""

    def __init__(
        self,
        class_counts: torch.Tensor,
        delta: float,
        reduction: Literal["mean", "sum"] = "mean",
    ):
        super(MarginBasedCrossEntropyLoss, self).__init__()
        self.class_counts = class_counts.float()
        self.delta = delta
        self.margins = self._compute_margins()
        self.reduction = reduction

    def _compute_margins(self) -> torch.Tensor:
        min_sqrt_n = torch.min(self.class_counts.pow(0.25))
        return self.delta * (min_sqrt_n / self.class_counts.pow(0.25))

    def forward(self, logits: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        margins = self.margins.to(logits.device)
        target_margins = margins[targets]
        logits_adjusted = logits.clone()
        logits_adjusted[torch.arange(logits.shape[0]), targets] -= target_margins
        return F.cross_entropy(logits_adjusted, targets, reduction=self.reduction)

    def __str__(self, indent: int = 0) -> str:
        indent_str = " " * indent
        return (
            f"{indent_str}MarginBasedCrossEntropyLoss(delta={self.delta},"
            + f"reduction={self.reduction})"
        )

    def __repr__(self, indent: int = 0) -> str:
        return self.__str__(indent=indent)


class GEARClient(Client):

    def __init__(
        self,
        index: int,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: Module,  # not used
        local_epochs: int,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        eps: float = 0.1,
        delta: float = 0.01,
        alpha: float = 2.0 / 255,
        adv_iters: int = 10,
        **kwargs,
    ):
        self.sample_per_class = torch.zeros(train_set.num_labels)
        uniq_val, uniq_count = np.unique(train_set.tensors[1], return_counts=True)
        for i, c in enumerate(uniq_val.tolist()):
            self.sample_per_class[c] = uniq_count[i]

        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            loss_fn=MarginBasedCrossEntropyLoss(self.sample_per_class, delta),
            optimizer_cfg=optimizer_cfg,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.hyper_params.update(eps=eps, delta=delta, alpha=alpha, adv_iters=adv_iters)

    def generate_adversarial(
        self,
        model: Module,
        inputs: torch.Tensor,
        targets: torch.Tensor,
        eps: float = 0.1,
        alpha: float = 2.0 / 255,
        iters: int = 10,
    ) -> torch.Tensor:
        """
        Generates adversarial examples using Projected Gradient Descent (PGD).

        Args:
            model (nn.Module): The neural network model.
            inputs (Tensor): Input samples.
            targets (Tensor): True labels for the inputs.
            alpha (float): Step size for each iteration.
            iters (int): Number of attack iterations.

        Returns:
            Tensor: Adversarial examples.
        """
        adv_inputs = inputs.clone().detach().requires_grad_(True)
        for _ in range(iters):
            outputs = model(adv_inputs)
            loss = F.cross_entropy(outputs, targets)
            model.zero_grad()
            loss.backward()
            perturbation = alpha * adv_inputs.grad.sign()
            adv_inputs = adv_inputs + perturbation
            adv_inputs = torch.min(torch.max(adv_inputs, inputs - eps), inputs + eps)
            adv_inputs = torch.clamp(adv_inputs, 0, 1).detach().requires_grad_(True)
        return adv_inputs.requires_grad_(False)

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (
            override_local_epochs if override_local_epochs > 0 else self.hyper_params.local_epochs
        )

        self.model.train()
        self.model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                X_adv = self.generate_adversarial(self.model, X, y)
                self.optimizer.zero_grad()
                y_hat = self.model(X_adv)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        running_loss /= epochs * len(self.train_set)
        self.model.cpu()
        clear_cuda_cache()
        return running_loss


class GEAR(CentralizedFL):

    def get_client_class(self):
        return GEARClient



"""`fluke` command line interface."""

import os
import sys
import uuid
from pathlib import Path
from typing import Any, List, Optional

import typer
from hydra import compose, initialize_config_dir
from omegaconf import DictConfig, OmegaConf
from rich.console import Console

sys.path.append(".")

from . import __version__  # NOQA
from .config import Configuration, ConfigurationError, OptimizerConfigurator  # NOQA
from .utils import get_class_from_qualified_name, get_loss, get_model, plot_distribution  # NOQA

console = Console()
app = typer.Typer()


def fluke_banner() -> None:
    from rich.panel import Panel

    fluke_pretty = run.__doc__
    console.print(Panel(fluke_pretty, subtitle=f"v{__version__}", style="bold white"), width=53)


def version_callback(value: bool) -> None:
    if value:
        print(f"fluke: {__version__}")
        raise typer.Exit()


def _compose_config(cfg_base: str, overrides: Optional[List[str]]) -> DictConfig:
    abs_config_path = (Path.cwd() / cfg_base).resolve()
    cfg_base_folder = abs_config_path.parent
    cfg_base_name = abs_config_path.stem
    with initialize_config_dir(
        config_dir=str(cfg_base_folder), job_name="fluke_cli", version_base=None
    ):
        return compose(config_name=cfg_base_name, overrides=overrides)


@app.command()
def centralized(
    exp_cfg: str = typer.Argument(..., help="Configuration file"),
    alg_cfg: str = typer.Argument(..., help="Config file for the algorithm to run"),
    epochs: int = typer.Option(0, help="Number of epochs to run"),
) -> None:
    """Run a centralized learning experiment."""

    from rich.progress import track
    from torch.optim import SGD

    from . import FlukeENV  # NOQA
    from .data import FastDataLoader  # NOQA
    from .data.datasets import Datasets  # NOQA
    from .evaluation import ClassificationEval  # NOQA
    from .utils.log import get_logger  # NOQA

    cfg = Configuration(exp_cfg, alg_cfg)
    FlukeENV().configure(cfg)
    data_container = Datasets.get(**cfg.data.dataset)

    device = FlukeENV().get_device()
    train_loader = FastDataLoader(
        *data_container.train,
        batch_size=cfg.client.batch_size,
        num_labels=data_container.num_classes,
        shuffle=True,
    )
    test_loader = FastDataLoader(
        *data_container.test, batch_size=10, num_labels=data_container.num_classes, shuffle=False
    )

    hp = cfg.method.hyperparameters
    model = get_model(mname=hp.model, **hp.net_args if "net_args" in hp else {})
    if "name" not in hp.client.optimizer:
        hp.client.optimizer.name = SGD
    optimizer_cfg = OptimizerConfigurator(
        optimizer_cfg=hp.client.optimizer, scheduler_cfg=hp.client.scheduler
    )
    optimizer, scheduler = optimizer_cfg(model)
    criterion = get_loss(hp.client.loss)
    evaluator = ClassificationEval(
        eval_every=cfg.eval.eval_every, n_classes=data_container.num_classes
    )
    history = []

    model.to(device)
    epochs = (
        epochs if epochs > 0 else int(max(1, cfg.protocol.n_rounds * cfg.protocol.eligible_perc))
    )

    exp_id = uuid.uuid4().hex
    exp_name = f"Centralized [{exp_id}]"
    log = get_logger(cfg.logger.name, name=exp_name, **cfg.logger.exclude("name"))
    log.init(**cfg, exp_id=exp_id)
    log.log(f"Centralized Learning [ #Epochs = {epochs} ]\n")

    for e in range(epochs):
        model.train()
        console.print(f"Epoch {e + 1}")
        for _, (X, y) in track(
            enumerate(train_loader), total=train_loader.n_batches, transient=True
        ):
            X, y = X.to(device), y.to(device)
            optimizer.zero_grad()
            y_hat = model(X)
            loss = criterion(y_hat, y)
            loss.backward()
            optimizer.step()
        scheduler.step()

        epoch_eval = evaluator.evaluate(e + 1, model, test_loader, criterion, device=device)
        history.append(epoch_eval)
        log.add_scalars(f"Epoch {e + 1}", epoch_eval, e + 1)
        log.pretty_log(epoch_eval, title=f"Performance [Epoch {e + 1}]")
        console.print()
    model.cpu()


@app.command()
def federation(
    exp_cfg: str = typer.Argument(..., help="Configuration file"),
    alg_cfg: str = typer.Argument(..., help="Config file for the algorithm to run"),
    overrides: Optional[List[str]] = typer.Argument(
        None, help='Overrides for the configuration, e.g. "exp.seed=10"'
    ),
    resume: str = typer.Option(None, help="Path to the checkpoint file to load."),
) -> None:
    """Run a federated learning experiment."""

    try:

        if overrides is not None:
            overrides_exp = [v for v in overrides if not v.startswith("method.")]
            overrides_alg = [v for v in overrides if v.startswith("method.")]
            exp_cfg = _compose_config(exp_cfg, overrides_exp)
            alg_cfg = _compose_config(alg_cfg, overrides_alg)
            OmegaConf.set_struct(exp_cfg, False)
            alg_cfg = OmegaConf.create({"method": alg_cfg})
            cfg = Configuration.from_dict(OmegaConf.merge(exp_cfg, alg_cfg))
        else:
            cfg = Configuration(exp_cfg, alg_cfg)
    except ConfigurationError:
        exit(1)
    except Exception as e:
        raise e

    _run_federation(cfg, resume)


@app.command()
def sweep(
    exp_cfg: str = typer.Argument(..., help="Configuration file"),
    alg_cfgs: List[str] = typer.Argument(..., help="Config file(s) for the algorithm(s) to run"),
) -> None:
    """Run a battery of federated learning experiments with different configurations."""

    for alg_cfg in alg_cfgs:
        for cfg in Configuration.sweep(exp_cfg, alg_cfg):
            try:
                _run_federation(cfg)
            except Exception as e:
                # print the error with rich
                console.print(f"[yellow]Error with experiment: {cfg.verbose()}")
                console.print(f"[red]{e}")
                console.print_exception(max_frames=10)
                continue


def _run_federation(cfg: Configuration, resume: str | None = None) -> None:
    import yaml
    from rich.panel import Panel
    from rich.pretty import Pretty

    from . import FlukeENV  # NOQA
    from .data import DataSplitter  # NOQA
    from .data.datasets import Datasets  # NOQA
    from .evaluation import ClassificationEval  # NOQA
    from .utils.log import get_logger  # NOQA

    FlukeENV().configure(cfg)
    data_container = Datasets.get(**cfg.data.dataset)
    evaluator = ClassificationEval(
        eval_every=cfg.eval.eval_every, n_classes=data_container.num_classes
    )
    FlukeENV().set_evaluator(evaluator)

    data_splitter = DataSplitter(
        dataset=data_container,
        distribution=cfg.data.distribution.name,
        dist_args=cfg.data.distribution.exclude("name"),
        **cfg.data.exclude("dataset", "distribution"),
    )

    fl_algo_class = get_class_from_qualified_name(cfg.method.name)
    fl_algo = fl_algo_class(cfg.protocol.n_clients, data_splitter, cfg.method.hyperparameters)

    if cfg.save and cfg.save.path:
        path = f"{cfg.save.path}_{fl_algo.id}"
        if not os.path.exists(path):
            os.makedirs(path)
        yaml.dump(cfg.to_dict(), open(f"{path}/config.yaml", "w"))

    log_name = f"{fl_algo.__class__.__name__} [{fl_algo.id}]"
    log = get_logger(cfg.logger.name, name=log_name, **cfg.logger.exclude("name"))
    log.init(**cfg, exp_id=fl_algo.id)

    fl_algo.set_callbacks([log])
    FlukeENV().set_logger(log)
    console.print(Panel(Pretty(fl_algo), title="FL algorithm", width=100))

    if resume is not None:
        fl_algo.load(resume)

    try:
        fl_algo.run(cfg.protocol.n_rounds, cfg.protocol.eligible_perc)
    except Exception as e:
        log.log(f"Error: {e}")
        FlukeENV().force_close()
        FlukeENV.clear()
        log.close()
        FlukeENV().close_cache()
        raise e

    log.close()


@app.command()
def clients_only(
    exp_cfg: str = typer.Argument(..., help="Configuration file"),
    alg_cfg: str = typer.Argument(..., help="Config file for the algorithm to run"),
    epochs: int = typer.Option(0, help="Number of epochs to run"),
) -> None:
    """Run a local training (for all clients) experiment."""

    import numpy as np
    from pandas import DataFrame
    from rich.progress import track
    from torch.optim import SGD

    # sys.path.append(".")
    from . import FlukeENV  # NOQA
    from .data import DataSplitter  # NOQA
    from .data.datasets import Datasets  # NOQA
    from .evaluation import ClassificationEval  # NOQA
    from .utils.log import get_logger  # NOQA

    cfg = Configuration(exp_cfg, alg_cfg)
    FlukeENV().configure(cfg)

    data_container = Datasets.get(**cfg.data.dataset)
    data_splitter = DataSplitter(
        dataset=data_container,
        distribution=cfg.data.distribution.name,
        dist_args=cfg.data.distribution.exclude("name"),
        **cfg.data.exclude("dataset", "distribution"),
    )

    device = FlukeENV().get_device()

    hp = cfg.method.hyperparameters
    if "name" not in hp.client.optimizer:
        hp.client.optimizer.name = SGD

    (clients_tr_data, clients_te_data), shared_test = data_splitter.assign(
        cfg.protocol.n_clients, hp.client.batch_size
    )

    criterion = get_loss(hp.client.loss)
    client_local_evals = []
    client_shared_evals = []
    if epochs == 0:
        epochs = int(cfg.protocol.n_rounds * hp.client.local_epochs * cfg.protocol.eligible_perc)
    cfg.exp_id = uuid.uuid4().hex
    exp_name = f"Clients-only [{cfg.exp_id}]"

    if "tags" in cfg.logger:
        cfg.logger.tags = [f"{cfg.exp_id}"] + cfg.logger.tags
    elif cfg.logger.name in ["WandBLog", "ClearMLLog"]:
        cfg.logger.tags = [f"{cfg.exp_id}"]

    # running_local_evals = {c: [] for c in range(cfg.protocol.n_clients)}
    # running_shared_evals = {c: [] for c in range(cfg.protocol.n_clients)}
    for i, (train_loader, test_loader) in enumerate(zip(clients_tr_data, clients_te_data)):
        name = f"Client [{i}]_{cfg.exp_id}"
        if cfg.logger.name == "WandBLog":
            cfg.logger.group = exp_name
            cfg.logger.reinit = True

        log = get_logger(cfg.logger.name, name=name, **cfg.logger.exclude("name"))
        log.init(**cfg)

        log.log(f"Client [{i + 1}/{cfg.protocol.n_clients}]")
        model = get_model(mname=hp.model, **hp.net_args if "net_args" in hp else {})
        model.to(device)
        optimizer_cfg = OptimizerConfigurator(
            optimizer_cfg=hp.client.optimizer, scheduler_cfg=hp.client.scheduler
        )
        optimizer, scheduler = optimizer_cfg(model)
        evaluator = ClassificationEval(
            eval_every=cfg.eval.eval_every, n_classes=data_container.num_classes
        )
        for e in track(range(epochs), description="Training...", transient=True):
            model.to(device)
            model.train()
            for _, (X, y) in enumerate(train_loader):
                X, y = X.to(device), y.to(device)
                optimizer.zero_grad()
                y_hat = model(X)
                loss = criterion(y_hat, y)
                loss.backward()
                optimizer.step()
            scheduler.step()

            if test_loader is not None:
                client_local_eval = evaluator.evaluate(
                    e + 1, model, test_loader, criterion, device=device
                )
                client_local_eval["epoch"] = e + 1
                log.add_scalars(f"Client[{i}].local_test", client_local_eval, e + 1)
                # running_local_evals[i].append(client_local_eval)
            if shared_test is not None:
                client_shared_eval = evaluator.evaluate(
                    e + 1, model, shared_test, criterion, device=device
                )
                client_shared_eval["epoch"] = e + 1
                log.add_scalars(f"Client[{i}].shared_test", client_shared_eval, e + 1)
                # running_shared_evals[i].append(client_shared_eval)

        perf = {}
        if test_loader is not None:
            perf["local_test"] = client_local_eval
            client_local_evals.append(client_local_eval)
        if shared_test is not None:
            perf["shared_test"] = client_shared_eval
            client_shared_evals.append(client_shared_eval)

        log.pretty_log(perf, title=f"Client [{i}] Performance")
        model.cpu()
        log.close()

    client_mean = {}
    if test_loader is not None:
        client_mu = DataFrame(client_local_evals).mean(numeric_only=True).to_dict()
        client_mean["local_test"] = {k: float(np.round(float(v), 5)) for k, v in client_mu.items()}
    if shared_test is not None:
        client_mu = DataFrame(client_shared_evals).mean(numeric_only=True).to_dict()
        client_mean["shared_test"] = {k: float(np.round(float(v), 5)) for k, v in client_mu.items()}

    log.pretty_log(client_mean, title="Overall local performance")


@app.callback()
def run(
    version: bool = typer.Option(
        None, "--version", help="Show the installed version of fluke", callback=version_callback
    )
) -> None:
    """\
\b\
    ██████  ████             █████
   ███░░███░░███            ░░███
  ░███ ░░░  ░███  █████ ████ ░███ █████  ██████
 ███████    ░███ ░░███ ░███  ░███░░███  ███░░███
░░░███░     ░███  ░███ ░███  ░██████░  ░███████
  ░███      ░███  ░███ ░███  ░███░░███ ░███░░░
  █████     █████ ░░████████ ████ █████░░██████
 ░░░░░     ░░░░░   ░░░░░░░░ ░░░░ ░░░░░  ░░░░░░\
    """

    fluke_banner()


def main() -> Any:
    return app()


if __name__ == "__main__":
    main()



"""This module contains the definition of the evaluation classes used to perform the evaluation
of the model client-side and server-side."""

import sys
from abc import ABC, abstractmethod
from typing import Any, Collection, Optional, Union, Literal

from pandas import DataFrame
import numpy as np
import torch
from torch.nn import Module
from torchmetrics import Accuracy, F1Score, Metric, Precision, Recall

sys.path.append(".")
sys.path.append("..")

from .data import FastDataLoader  # NOQA

__all__ = ["Evaluator", "ClassificationEval", "PerformanceTracker"]


class Evaluator(ABC):
    """This class is the base class for all evaluators in :mod:`fluke`.
    An evaluator object should be used to perform the evaluation of a (federated) model.

    Args:
        eval_every (int): The evaluation frequency expressed as the number of rounds between
            two evaluations. Defaults to 1, i.e., evaluate the model at each round.

    Attributes:
        eval_every (int): The evaluation frequency.
    """

    def __init__(self, eval_every: int = 1):
        self.eval_every: int = eval_every

    @abstractmethod
    def evaluate(
        self,
        round: int,
        model: Module,
        eval_data_loader: FastDataLoader,
        loss_fn: Optional[torch.nn.Module],
        additional_metrics: Optional[dict[str, Metric]] = None,
        **kwargs,
    ) -> dict[str, Any]:
        """Evaluate the model.

        Args:
            round (int): The current
            model (Module): The model to evaluate.
            eval_data_loader (FastDataLoader): The data loader to use for evaluation.
            loss_fn (torch.nn.Module, optional): The loss function to use for evaluation.
            additional_metrics (dict[str, Metric], optional): Additional metrics to use for
                evaluation. If provided, they are added to the default metrics.
            **kwargs: Additional keyword arguments.

        Returns:
            dict[str, Any]: A dictionary containing the computed metrics.
        """
        raise NotImplementedError

    def __call__(
        self,
        round: int,
        model: Module,
        eval_data_loader: FastDataLoader,
        loss_fn: Optional[torch.nn.Module],
        additional_metrics: Optional[dict[str, Metric]] = None,
        **kwargs,
    ) -> dict:
        """Evaluate the model.

        Note:
            This method is equivalent to ``evaluate``.

        Args:
            round (int): The current round.
            model (Module): The model to evaluate.
            eval_data_loader (FastDataLoader): The data loader to use for evaluation.
            loss_fn (torch.nn.Module, optional): The loss function to use for evaluation.
            **kwargs: Additional keyword arguments.

        Returns:
            dict[str, Any]: A dictionary containing the computed metrics.
        """
        return self.evaluate(
            round=round,
            model=model,
            eval_data_loader=eval_data_loader,
            loss_fn=loss_fn,
            **kwargs,
        )


class ClassificationEval(Evaluator):
    """Evaluate a PyTorch model for classification.
    The metrics computed are ``accuracy``, ``precision``, ``recall``, ``f1`` and the loss according
    to the provided loss function ``loss_fn`` when calling the method ``evaluation``.
    Metrics are computed both in a micro and macro fashion.

    Args:
        eval_every (int): The evaluation frequency.
        n_classes (int): The number of classes.
        **metrics (dict[str, Metric]): The metrics to use for evaluation. If not provided, the
            default metrics are used: ``accuracy``, ``macro_precision``, ``macro_recall``,
            ``macro_f1``, ``micro_precision``, ``micro_recall`` and ``micro_f1``.

    Attributes:
        eval_every (int): The evaluation frequency.
        n_classes (int): The number of classes.
    """

    def __init__(self, eval_every: int, n_classes: int, **metrics: Metric):
        super().__init__(eval_every=eval_every)
        self.n_classes: int = n_classes

        self.metrics = {}

        # if kwargs is empty
        if not metrics:
            self.metrics = {
                "accuracy": Accuracy(task="multiclass", num_classes=self.n_classes, top_k=1),
                "macro_precision": Precision(
                    task="multiclass",
                    num_classes=self.n_classes,
                    top_k=1,
                    average="macro",
                ),
                "macro_recall": Recall(
                    task="multiclass",
                    num_classes=self.n_classes,
                    top_k=1,
                    average="macro",
                ),
                "macro_f1": F1Score(
                    task="multiclass",
                    num_classes=self.n_classes,
                    top_k=1,
                    average="macro",
                ),
                "micro_precision": Precision(
                    task="multiclass",
                    num_classes=self.n_classes,
                    top_k=1,
                    average="micro",
                ),
                "micro_recall": Recall(
                    task="multiclass",
                    num_classes=self.n_classes,
                    top_k=1,
                    average="micro",
                ),
                "micro_f1": F1Score(
                    task="multiclass",
                    num_classes=self.n_classes,
                    top_k=1,
                    average="micro",
                ),
            }
        else:
            self.metrics = metrics

    def add_metric(self, name: str, metric: Metric) -> None:
        """Add a metric to the evaluator.

        Args:
            name (str): The name of the metric.
            metric (Metric): The metric to add.
        """
        if name in self.metrics:
            raise ValueError(f"Metric {name} already exists.")
        self.metrics[name] = metric

    @torch.no_grad()
    def evaluate(
        self,
        round: int,
        model: torch.nn.Module,
        eval_data_loader: Union[FastDataLoader, Collection[FastDataLoader]],
        loss_fn: Optional[torch.nn.Module] = None,
        additional_metrics: Optional[dict[str, Metric]] = None,
        device: torch.device = torch.device("cpu"),
    ) -> dict:
        """Evaluate the model. The metrics computed are ``accuracy``, ``precision``, ``recall``,
        ``f1`` and the loss according to the provided loss function ``loss_fn``. Metrics are
        computed both in a micro and macro fashion.

        Warning:
            The loss function ``loss_fn`` should be defined on the same device as the model.
            Moreover, it is assumed that the only arguments of the loss function are the predicted
            values and the true values.

        Args:
            round (int): The current round.
            model (torch.nn.Module): The model to evaluate. If ``None``, the method returns an
                empty dictionary.
            eval_data_loader (Union[FastDataLoader, Collection[FastDataLoader]]):
                The data loader(s) to use for evaluation. If ``None``, the method returns an empty
                dictionary.
            loss_fn (torch.nn.Module, optional): The loss function to use for evaluation.
            additional_metrics (dict[str, Metric], optional): Additional metrics to use for
                evaluation. If provided, they are added to the default metrics.
            device (torch.device, optional): The device to use for evaluation. Defaults to "cpu".

        Returns:
            dict: A dictionary containing the computed metrics.
        """
        from .utils import clear_cuda_cache  # NOQA

        if (round != 1) and (round % self.eval_every != 0):
            return {}

        if (model is None) or (eval_data_loader is None):
            return {}

        model_device = torch.device("cpu")
        if next(model.parameters(), None) is not None:
            model_device = next(model.parameters()).device
        model.eval()
        model.to(device)
        losses = []
        matrics_values = {k: [] for k in self.metrics.keys()}
        loss, cnt = 0, 0

        if additional_metrics is None:
            additional_metrics = {}

        add_metric_values = {k: [] for k in additional_metrics.keys()}

        if not isinstance(eval_data_loader, list):
            eval_data_loader = [eval_data_loader]

        for data_loader in eval_data_loader:
            for metric in self.metrics.values():
                metric.reset()

            for metric in additional_metrics.values():
                metric.reset()

            loss = 0
            for X, y in data_loader:
                X, y = X.to(device), y.to(device)
                with torch.no_grad():
                    y_hat = model(X)
                    if loss_fn is not None:
                        loss += loss_fn(y_hat, y).item()

                for metric in self.metrics.values():
                    metric.update(y_hat.cpu(), y.cpu())

                if additional_metrics:
                    for metric in additional_metrics.values():
                        metric.update(y_hat.cpu(), y.cpu())

            cnt += len(data_loader)

            for k, v in self.metrics.items():
                matrics_values[k].append(v.compute().item())

            if additional_metrics:
                for k, v in additional_metrics.items():
                    add_metric_values[k].append(v.compute().item())

            losses.append(loss / cnt)

        model.to(model_device)
        clear_cuda_cache()

        result = {m: np.round(sum(v) / len(v), 5).item() for m, v in matrics_values.items()}
        result.update(
            {m: np.round(sum(v) / len(v), 5).item() for m, v in add_metric_values.items()}
        )

        if loss_fn is not None:
            result["loss"] = np.round(sum(losses) / len(losses), 5).item()

        return result

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}(eval_every={self.eval_every}"
            + f", n_classes={self.n_classes})[{', '.join(self.metrics.keys())}]"
        )

    def __repr__(self) -> str:
        return str(self)


def _compute_mean(evals: dict[str, float]) -> dict[str, float]:
    df_data = DataFrame(evals.values())
    client_mean = df_data.mean(numeric_only=True).to_dict()
    return {k: float(np.round(float(v), 5)) for k, v in client_mean.items()}


class PerformanceTracker:

    def __init__(self):

        self._performance = {
            "global": {},  # round -> evals
            "locals": {},  # round -> {client_id -> evals}
            "pre-fit": {},  # round -> {client_id -> evals}
            "post-fit": {},  # round -> {client_id -> evals}
            "comm": {0: 0},  # round -> comm_cost
            "mem": {},  # round -> mem_usage
        }

    def add(
        self,
        perf_type: Literal["global", "locals", "pre-fit", "post-fit", "comm", "mem"],
        metrics: dict[str, float] | float,
        round: int = 0,
        client_id: Optional[int] = None,
    ) -> None:
        """Add performance metrics for a specific type and client.

        Args:
            perf_type (Literal["global", "locals", "pre-fit", "post-fit", "comm", "mem"]):
                The type of performance metrics to add.
            metrics (dict[str, float] | float): The performance metrics to add. If `perf_type`
                is "comm", this should be a single float value representing the communication cost.
            round (int, optional): The current round. Defaults to 0.
            client_id (int, optional): The client ID for local performance metrics. Defaults to
                None for global metrics.
        """
        if perf_type not in self._performance:
            raise ValueError(f"Unknown performance type: {perf_type}")

        if perf_type == "comm" or perf_type == "mem":
            if not isinstance(metrics, (float, int)):
                raise ValueError(f"Metrics for {perf_type} must be a float, got {type(metrics)}")
            if round not in self._performance[perf_type]:
                self._performance[perf_type][round] = metrics
            else:
                self._performance[perf_type][round] += metrics
        else:
            if round not in self._performance[perf_type]:
                self._performance[perf_type][round] = {}
            if perf_type != "global":
                self._performance[perf_type][round][client_id] = metrics
            else:
                self._performance[perf_type][round] = metrics

    def get(
        self,
        perf_type: Literal["global", "locals", "pre-fit", "post-fit", "comm", "mem"],
        round: int,
    ) -> Union[dict, float]:
        """Get performance metrics for a specific type and round.

        Args:
            perf_type (Literal["global", "locals", "pre-fit", "post-fit", "comm", "mem"]):
                The type of performance metrics to retrieve.
            round (int): The round for which to retrieve the metrics.

        Raises:
            ValueError: If the `perf_type` is unknown.

        Returns:
            Union[dict, float]: The performance metrics for the specified type and round.
                If `perf_type` is "comm" or "mem", returns a float; otherwise, returns a dict.
        """
        if perf_type not in self._performance:
            raise ValueError(f"Unknown performance type: {perf_type}")

        if round not in self._performance[perf_type]:
            return {} if perf_type not in ["comm", "mem"] else 0.0

        if perf_type in ["comm", "mem"]:
            return self._performance[perf_type][round]
        else:
            return self._performance[perf_type][round].copy()

    def __getitem__(self, item: str) -> dict:
        """Get performance metrics for a specific type.

        Args:
            item (str): The type of performance metrics to retrieve.

        Raises:
            ValueError: If the `item` is unknown.

        Returns:
            dict: The performance metrics for the specified type.
        """
        if item not in self._performance:
            raise ValueError(f"Unknown performance type: {item}")
        return self._performance[item].copy()

    def summary(
        self,
        perf_type: Literal["global", "locals", "pre-fit", "post-fit", "comm", "mem"],
        round: int,
        include_round: bool = True,
        force_round: bool = True,
    ) -> Union[dict, float]:
        """Get the summary of the performance metrics for a specific type.

        Summary metrics are computed as the mean of the metrics for the specified type
        and round. If `perf_type` is "comm", the total communication cost is returned.
        If `perf_type` is "mem", the memory usage is returned.
        If `perf_type` is "global", the metrics are returned as they are.

        Args:
            perf_type (Literal["global", "locals", "pre-fit", "post-fit", "comm", "mem"]):
                The type of performance metrics to retrieve.
            round (int): The round for which to compute the summary of the metrics.
            include_round (bool, optional): Whether to include the round number in the returned
                metrics. Defaults to `True`.
            force_round (bool, optional): If `True`, the method will return the metrics for the
                specified round if it exists, otherwise it will return the metrics for the
                latest round. Defaults to `False`.

        Raises:
            ValueError: If the `perf_type` is unknown or if there are no metrics for the specified
                type and round.

        Returns:
            Union[dict, float]: The summary performance metrics for the specified type.
        """
        if perf_type not in self._performance:
            raise ValueError(f"Unknown performance type: {perf_type}")

        if not self._performance[perf_type]:
            return {} if perf_type not in ["comm", "mem"] else 0.0

        if force_round:
            the_round = max(self._performance[perf_type].keys())
        elif round not in self._performance[perf_type]:
            return {}
        else:
            the_round = round

        if perf_type == "mem":
            return self._performance[perf_type][the_round]

        if perf_type == "comm":
            return sum(list(self._performance[perf_type].values()))

        if perf_type == "global":
            metrics = self._performance[perf_type][the_round].copy()
        else:
            metrics = _compute_mean(self._performance[perf_type][the_round])
            metrics["support"] = len(self._performance[perf_type][the_round])

        if include_round and metrics:
            metrics["round"] = the_round
        return metrics



"""Implementation of the APFL [APFL20]_ algorithm.

References:
    .. [APFL20] Yuyang Deng, Mohammad Mahdi Kamani, and Mehrdad Mahdavi. Adaptive Personalized
       Federated Learning. In arXiv (2020). URL: https://arxiv.org/abs/2003.13461
"""

import sys
from typing import Collection, Sequence

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import PersonalizedFL  # NOQA
from ..client import Client, PFLClient  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import clear_cuda_cache  # NOQA
from ..utils.model import merge_models  # NOQA

__all__ = ["APFLClient", "APFLServer", "APFL"]


class APFLClient(PFLClient):

    def __init__(
        self,
        index: int,
        model: torch.nn.Module,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: Module,
        local_epochs: int = 3,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        lam: float = 0.25,
        **kwargs,
    ):
        super().__init__(
            index=index,
            model=model,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.hyper_params.update(lam=lam)

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (
            override_local_epochs if override_local_epochs > 0 else self.hyper_params.local_epochs
        )

        self.model.train()
        self.personalized_model.train()

        self.model.to(self.device)
        self.personalized_model.to(self.device)

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        if self.pers_optimizer is None:
            self.pers_optimizer, self.pers_scheduler = self._optimizer_cfg(self.personalized_model)

        running_loss = 0.0
        for _ in range(epochs):
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)

                # Global
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()

                # Local
                self.pers_optimizer.zero_grad()
                y_hat = merge_models(self.model, self.personalized_model, self.hyper_params.lam)(X)
                local_loss = self.hyper_params.loss_fn(y_hat, y)
                local_loss.backward()
                self._clip_grads(self.personalized_model)
                self.pers_optimizer.step()

            self.scheduler.step()
            self.pers_scheduler.step()

        running_loss /= epochs * len(self.train_set)
        self.model.cpu()
        self.personalized_model.cpu()
        clear_cuda_cache()

        self.personalized_model = merge_models(
            self.model, self.personalized_model, self.hyper_params.lam
        )

        return running_loss


class APFLServer(Server):

    def __init__(
        self,
        model: Module,
        test_set: FastDataLoader,
        clients: Sequence[Client],
        weighted: bool = False,
        tau: int = 3,
        **kwargs,
    ):
        super().__init__(model=model, test_set=test_set, clients=clients, weighted=weighted)
        self.hyper_params.update(tau=tau)

    @torch.no_grad()
    def aggregate(self, eligible: Sequence[Client], client_models: Collection[Module]) -> None:
        """Aggregate the models of the eligible clients every `hyper_params.tau` rounds.

        Args:
            eligible (Sequence[Client]): The clients that are eligible to participate in the
                aggregation.
            client_models (Collection[Module]): The models of the clients to aggregate.
        """
        if self.rounds % self.hyper_params.tau != 0:
            # Ignore the sent models and clear the channel's cache
            self.channel.clear(self)
        else:
            super().aggregate(eligible, client_models)


class APFL(PersonalizedFL):

    def get_client_class(self) -> type[PFLClient]:
        return APFLClient

    def get_server_class(self) -> type[Server]:
        return APFLServer



"""
Utility module for `fluke.distr`.
"""

from torch.nn import Module

from ..utils import get_model

__all__ = ["ModelBuilder"]


class ModelBuilder:
    """A class to build and return instances of a model class with specified arguments.

    This class is useful for creating model instances in a flexible way, allowing for
    different configurations of the model without directly instantiating it.

    Attributes:
        model_class (type): The class of the model to be instantiated.
        args (tuple): Positional arguments to pass to the model constructor.
        kwargs (dict): Keyword arguments to pass to the model constructor.
    """

    def __init__(self, model_class: type, **kwargs):
        self.model_class = model_class
        self.kwargs = kwargs

    def build(self) -> Module:
        """Build and return an instance of the model.

        Returns:
            Module: An instance of the model class initialized with the provided arguments.
        """
        return get_model(mname=self.model_class, **self.kwargs)



"""This module contains utility functions and classes used in ``fluke``."""

from __future__ import annotations

import importlib
import os
import sys
import warnings
from typing import TYPE_CHECKING, Any, Collection, Literal, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import psutil
import seaborn as sns
import torch
from sklearn.model_selection import train_test_split
from torch.nn import Module
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler

sys.path.append(".")
sys.path.append("..")

if TYPE_CHECKING:
    from client import Client  # NOQA
    from server import Server  # NOQA

from .. import FlukeCache, FlukeENV, custom_formatwarning  # NOQA

warnings.formatwarning = custom_formatwarning

__all__ = [
    "log",
    "model",
    "ClientObserver",
    "ServerObserver",
    "bytes2human",
    "cache_obj",
    "clear_cuda_cache",
    "get_class_from_str",
    "get_class_from_qualified_name",
    "get_full_classname",
    "get_loss",
    "get_model",
    "get_optimizer",
    "get_scheduler",
    "flatten_dict",
    "import_module_from_str",
    "memory_usage",
    "plot_distribution",
    "retrieve_obj",
    "safe_train_test_split",
]


class ClientObserver:
    """Client observer interface.
    This interface is used to observe the client during the federated learning process.
    For example, it can be used to log the performance of the local model, as it is done by the
    :class:`fluke.utils.log.Log` class.
    """

    def start_fit(self, round: int, client_id: int, model: Module, **kwargs):
        """This method is called when the client starts the local training process.

        Args:
            round (int): The round number.
            client_id (int):  The client ID.
            model (Module): The local model before training.
            **kwargs (dict): Additional keyword arguments.
        """
        pass

    def end_fit(self, round: int, client_id: int, model: Module, loss: float, **kwargs):
        """This method is called when the client ends the local training process.

        Args:
            round (int): The round number.
            client_id (int): The client ID.
            model (Module): The local model after the local training.
            loss (float): The average loss incurred by the local model during training.
            **kwargs (dict): Additional keyword arguments.
        """
        pass

    def client_evaluation(
        self,
        round: int,
        client_id: int,
        phase: Literal["pre-fit", "post-fit"],
        evals: dict[str, float],
        **kwargs,
    ):
        """This method is called when the client evaluates the local model.
        The evaluation can be done before ('pre-fit') and/or after ('post-fit') the local
        training process. The 'pre-fit' evaluation is usually the evaluation of the global model on
        the local test set, and the 'post-fit' evaluation is the evaluation of the just updated
        local model on the local test set.

        Args:
            round (int): The round number.
            client_id (int): The client ID.
            phase (Literal['pre-fit', 'post-fit']): Whether the evaluation is done before or after
                the local training process.
            evals (dict[str, float]): The evaluation results.
            **kwargs (dict): Additional keyword arguments.
        """
        pass

    def track_item(self, round: int, client_id: int, item: str, value: float, **kwargs) -> None:
        """This method is called when the client aims to log an item.

        Args:
            round (int): The round number.
            client_id (int): The client ID.
            item (str): The name of the log item.
            value (float): The value of the log item.
            **kwargs (dict): Additional keyword arguments.
        """
        pass


class ServerObserver:
    """Server observer interface.
    This interface is used to observe the server during the federated learning process.
    For example, it can be used to log the performance of the global model and the communication
    costs, as it is done by the ``Log`` class.
    """

    def start_round(self, round: int, global_model: Any) -> None:
        """This method is called when a new round starts.

        Args:
            round (int): The round number.
            global_model (Any): The current global model.
        """
        pass

    def end_round(self, round: int) -> None:
        """This method is called when a round ends.

        Args:
            round (int): The round number.
        """
        pass

    def selected_clients(self, round: int, clients: Collection) -> None:
        """This method is called when the clients have been selected for the current round.

        Args:
            round (int): The round number.
            clients (Collection): The clients selected for the current round.
        """
        pass

    def server_evaluation(
        self,
        round: int,
        eval_type: Literal["global", "locals"],
        evals: Union[dict[str, float], dict[int, dict[str, float]]],
        **kwargs,
    ) -> None:
        """This method is called when the server evaluates the global or the local models on its
        test set.

        Args:
            round (int): The round number.
            eval_type (Literal['global', 'locals']): The type of evaluation. If 'global', the
                evaluation is done on the global model. If 'locals', the evaluation is done on the
                local models of the clients on the test set of the server.
            evals (dict[str, float] | dict[int, dict[str, float]]): The evaluation metrics. In case
                of 'global' evaluation, it is a dictionary with the evaluation metrics. In case of
                'locals' evaluation, it is a dictionary of dictionaries where the keys are the
                client IDs and the values are the evaluation metrics.
        """
        pass

    def finished(self, round: int) -> None:
        """This method is called when the federated learning process has ended.

        Args:
            round (int): The last round number.
        """
        pass

    def interrupted(self) -> None:
        """This method is called when the federated learning process has been interrupted."""
        pass

    def early_stop(self, round: int) -> None:
        """This method is called when the federated learning process has been stopped due to an
        early stopping criterion.

        Args:
            round (int): The last round number.
        """
        pass

    def track_item(self, round: int, item: str, value: float) -> None:
        """This method is called when the server aims log an item.

        Args:
            round (int): The round number.
            item (str): The name of the log item.
            value (float): The value of the log item.
        """
        pass


def safe_train_test_split(
    X: torch.Tensor, y: torch.Tensor, test_size: float, client_id: int | None = None
) -> tuple[torch.Tensor, Optional[torch.Tensor], torch.Tensor, Optional[torch.Tensor]]:
    try:
        if test_size == 0.0:
            return X, None, y, None
        else:
            return train_test_split(X, y, test_size=test_size, stratify=y)
    except ValueError:
        client_str = f"[Client {client_id}]" if client_id is not None else ""
        warnings.warn(f"Stratified split failed for {client_str}. Falling back to random split.")
        return train_test_split(X, y, test_size=test_size)


def import_module_from_str(name: str) -> Any:
    """Import a module from its name.

    Args:
        name (str): The name of the module.

    Returns:
        Any: The module.
    """
    components = name.split(".")
    mod = importlib.import_module(".".join(components[:-1]))
    mod = getattr(mod, components[-1])
    return mod


def get_class_from_str(module_name: str, class_name: str) -> Any:
    """Get a class from its name.
    This function is used to get a class from its name and the name of the module where it is
    defined. It is used to dynamically import classes.

    Args:
        module_name (str): The name of the module where the class is defined.
        class_name (str): The name of the class.

    Returns:
        Any: The class.
    """
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name)
    return class_


def get_class_from_qualified_name(qualname: str) -> Any:
    """Get a class from its fully qualified name.

    Args:
        qualname (str): The fully qualified name of the class.

    Returns:
        Any: The class.
    """
    module_name = ".".join(qualname.split(".")[:-1])
    class_name = qualname.split(".")[-1]
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name)
    return class_


def get_loss(lname: str) -> Module:
    """Get a loss function from its name.
    The supported loss functions are the ones defined in the ``torch.nn`` module.

    Args:
        lname (str): The name of the loss function.

    Returns:
        Module: The loss function.
    """
    return get_class_from_str("torch.nn", lname)()


def get_model(mname: str, **kwargs) -> Module:
    """Get a model from its name.
    This function is used to get a torch model from its name and the name of the module where it is
    defined. It is used to dynamically import models. If ``mname`` is not a fully qualified name,
    the model is assumed to be defined in the :mod:`fluke.nets` module.

    Args:
        mname (str): The name of the model.
        **kwargs: The keyword arguments to pass to the model's constructor.

    Returns:
        Module: The model.
    """
    if "." in mname:
        module_name = ".".join(mname.split(".")[:-1])
        mname = mname.split(".")[-1]
    else:
        module_name: str = "fluke.nets"
    return get_class_from_str(module_name, mname)(**kwargs)


def get_full_classname(classtype: type) -> str:
    """Get the fully qualified name of a class.

    Args:
        classtype (type): The class.

    Returns:
        str: The fully qualified name of the class.

    Example:
        Let ``A`` be a class defined in the module ``fluke.utils``

        .. code-block:: python
            :linenos:

            # This is the content of the file fluke/utils.py
            class A:
                pass

            get_full_classname(A) # 'fluke.utils.A'

        If the class is defined in the ``__main__`` module, then:

        .. code-block:: python
            :linenos:

            if __name__ == "__main__":
                class B:
                    pass

                get_full_classname(B) # '__main__.B'
    """
    return f"{classtype.__module__}.{classtype.__name__}"


def get_optimizer(oname: str) -> type[Optimizer]:
    """Get an optimizer from its name.
    This function is used to get an optimizer from its name. It is used to dynamically import
    optimizers. The supported optimizers are the ones defined in the :mod:`torch.optim` module.

    Args:
        oname (str): The name of the optimizer.

    Returns:
        type[Optimizer]: The optimizer class.
    """
    return get_class_from_str("torch.optim", oname)


def get_scheduler(sname: str) -> type[LRScheduler]:
    """Get a learning rate scheduler from its name.
    This function is used to get a learning rate scheduler from its name. It is used to dynamically
    import learning rate schedulers. The supported schedulers are the ones defined in the
    :mod:`torch.optim.lr_scheduler` module.

    Args:
        sname (str): The name of the scheduler.

    Returns:
        torch.nn.Module: The learning rate scheduler.
    """
    return get_class_from_str("torch.optim.lr_scheduler", sname)


def clear_cuda_cache(ipc: bool = False):
    """Clear the CUDA cache. This function should be used to free the GPU memory after the training
    process has ended. It is usually used after the local training of the clients.

    Args:
        ipc (bool, optional): Whether to force collecting GPU memory after it has been released by
            CUDA IPC.
    """
    torch.cuda.empty_cache()
    if ipc and torch.cuda.is_available():
        torch.cuda.ipc_collect()


def plot_distribution(clients: list[Client], train: bool = True, plot_type: str = "ball") -> None:
    """Plot the distribution of classes for each client.
    This function is used to plot the distribution of classes for each client. The plot can be a
    scatter plot, a heatmap, or a bar plot. The scatter plot (``plot_type='ball'``) shows filled
    circles whose size is proportional to the number of examples of a class. The heatmap
    (``plot_type='mat'``) shows a matrix where the rows represent the classes and the columns
    represent the clients with a color intensity proportional to the number of examples of a class.
    The bar plot (``plot_type='bar'``) shows a stacked bar plot where the height of the bars is
    proportional to the number of examples of a class.

    Warning:
        If the number of clients is greater than 30, the type is automatically switched to
        ``'bar'`` for better visualization.

    Args:
        clients (list[Client]): The list of clients.
        train (bool, optional): Whether to plot the distribution on the training set. If ``False``,
            the distribution is plotted on the test set. Defaults to ``True``.
        plot_type (str, optional): The type of plot. It can be ``'ball'``, ``'mat'``, or ``'bar'``.
            Defaults to ``'ball'``.
    """
    assert plot_type in [
        "bar",
        "ball",
        "mat",
    ], "Invalid plot type. Must be 'bar', 'ball' or 'mat'."
    if len(clients) > 30 and plot_type != "bar":
        warnings.warn("Too many clients to plot. Switching to 'bar' plot.")
        plot_type = "bar"

    client = {}
    for c in clients:
        client[c.index] = c.train_set.tensors[1] if train else c.test_set.tensors[1]

    # Count the occurrences of each class for each client
    class_counts = {
        client_idx: torch.bincount(client_data).tolist()
        for client_idx, client_data in enumerate(client.values())
    }

    # Maximum number of classes
    num_classes = max(len(counts) for counts in class_counts.values())

    _, ax = plt.subplots(figsize=(12, 6))
    ax.set_xticks(range(len(client)))

    class_matrix = np.zeros((num_classes, len(client)))
    for client_idx, counts in class_counts.items():
        for class_idx, count in enumerate(counts):
            class_matrix[class_idx, client_idx] = count
            # Adjusting size based on the count
            if plot_type == "ball":
                size = count * 1  # Adjust the scaling factor as needed
                ax.scatter(client_idx, class_idx, s=size, alpha=0.6)
                ax.set_yticks(range(num_classes))
                ax.text(
                    client_idx,
                    class_idx,
                    str(count),
                    va="center",
                    ha="center",
                    color="black",
                    fontsize=9,
                )
    plt.title("Number of Examples per Class for Each Client", fontsize=12)
    ax.grid(False)
    if plot_type == "mat":
        ax.set_yticks(range(num_classes))
        sns.heatmap(
            class_matrix,
            ax=ax,
            cmap="viridis",
            annot=class_matrix,
            fmt="g",
            cbar=False,
            annot_kws={"fontsize": 6},
        )
    elif plot_type == "bar":
        df = pd.DataFrame(
            class_matrix.T,
            index=[f"{i}" for i in range(len(client))],
            columns=[f"{i}" for i in range(num_classes)],
        )
        step = int(max(1, np.ceil(len(clients) / 50)))
        df.plot(
            ax=ax,
            kind="bar",
            stacked=True,
            xticks=[x for x in range(0, len(clients), step)],
            color=sns.color_palette("viridis", num_classes),
        ).legend(loc="upper right")
    plt.xlabel("clients")
    plt.ylabel("classes")
    plt.show()


def bytes2human(n: int) -> str:
    """Convert bytes to human-readable format.

    See:
        https://psutil.readthedocs.io/en/latest/#psutil.Process.memory_info

    Args:
        n (int): The number of bytes.

    Returns:
        str: The number of bytes in human-readable format.

    Example:
        .. code-block:: python
            :linenos:

            >>> bytes2human(10000)
            '9.8K'
            >>> bytes2human(100001221)
            '95.4M'
            >>> bytes2human(1024 ** 3)
            '1.0G'

    """
    symbols = ("K", "M", "G", "T", "P", "E", "Z", "Y")
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if abs(n) >= prefix[s]:
            value = float(n) / prefix[s]
            return "%.1f %sB" % (value, s)
    return "%s B" % n


def memory_usage() -> tuple[int, int, int]:
    """Get the memory usage of the current process.

    Returns:
        tuple[int, int, int]: The resident set size (RSS), the virtual memory size (VMS), and the
            current CUDA reserved memory. If the device is not a CUDA device, the reserved memory is
            0.
    """
    proc = psutil.Process(os.getpid())

    if FlukeENV().get_device().type == "cuda":
        cuda_device = torch.device(FlukeENV().get_device())
        current_reserved = torch.cuda.memory_reserved(cuda_device)
    else:
        current_reserved = 0
    return proc.memory_info().rss, proc.memory_info().vms, current_reserved


def retrieve_obj(key: str, party: Client | Server | None = None, pop: bool = True) -> Any:
    """Load an object from the cache (disk).
    If the object is not found in the cache, it returns ``None``.

    Warning:
        This method assumes the cache is already initialized and opened.
        If it is not, it will raise an exception.

    Args:
        key (str, optional): specific key to add to the default filename.
        party (Client | Server, optional): the party for which to load the object. Defaults to
            ``None``.
        pop (bool, optional): Whether to remove the object from the cache after loading it.
            Defaults to ``True``.

    Returns:
        Any: The object retrieved from the cache.
    """
    prefix = ""
    if party is not None:
        prefix = f"c{party.index}_" if hasattr(party, "index") else "server_"
    cache = FlukeENV().get_cache()
    obj = cache.pop(f"{prefix}{key}", copy=(party is None)) if pop else cache.get(f"{prefix}{key}")
    return obj


def cache_obj(
    obj: Any, key: str, party: Server | Client | None = None
) -> FlukeCache.ObjectRef | None:
    """Move the object from the RAM to the disk cache to free up memory.
    If the object is ``None``, it returns ``None`` without caching it.

    Warning:
        This method assumes the cache is already initialized and opened.
        If it is not, it will raise an exception.

    Args:
        obj (Any): The object to cache.
        key (str, optional): The key associated with the object.
        party (Client | Server): the party for which to unload the model. Defaults to ``None``.

    Returns:
        FlukeCache.ObjectRef: The object reference identifier. If the object is ``None``,
        it returns ``None``.
    """
    if obj is None:
        return None
    prefix = ""
    if party is not None:
        prefix = f"c{party.index}_" if hasattr(party, "index") else "server_"
    cache = FlukeENV().get_cache()
    return cache.push(f"{prefix}{key}", obj)


def _flatten_dict(d: dict, parent_key: str = "", sep: str = ".") -> dict:
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(_flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def flatten_dict(nested_dict: dict, sep: str = ".") -> dict:
    """Flatten a nested dictionary.
    The flatten dictionary is a dictionary where the keys are the concatenation of the keys of the
    nested dictionary separated by a separator.

    Args:
        nested_dict (dict): Nested dictionary.
        sep (str, optional): Separator. Defaults to '.'.

    Returns:
        dict: Flattened dictionary.

    Example:
        .. code-block:: python
            :linenos:

            d = {'a': 1, 'b': {'c': 2, 'd': {'e': 3}}}
            flatten_dict(d)
            # Output: {'a': 1, 'b.c': 2, 'b.d.e': 3}

    """
    return _flatten_dict(nested_dict, sep=sep)



"""Implementation of the Federated Averaging with Body Aggregation and Body Update [FedBABU22]_
algorithm.

References:
    .. [FedBABU22] Jaehoon Oh, Sangmook Kim, Se-Young Yun. FedBABU: Towards Enhanced Representation
       for Federated Image Classification. In ICLR (2022). URL: https://arxiv.org/abs/2106.06042
"""

import sys
from typing import Sequence

from rich.progress import Progress
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from .. import FlukeENV  # NOQA
from ..algorithms import PersonalizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderHeadNet  # NOQA
from ..server import Server  # NOQA
from ..utils import clear_cuda_cache  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA


class FedBABUClient(Client):

    def __init__(
        self,
        index: int,
        model: EncoderHeadNet,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: Module,
        local_epochs: int,
        mode: str,
        fine_tuning_epochs: int,
        clipping: float = 0,
        **kwargs,
    ):
        assert mode in ["head", "body", "full"]
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.hyper_params.update(mode=mode)
        self.model = model
        self._save_to_cache()

    def send_model(self) -> None:
        self.channel.send(Message(self.model.encoder, "model", self.index, inmemory=True), "server")

    def receive_model(self) -> None:
        msg = self.channel.receive(self.index, "server", msg_type="model")
        safe_load_state_dict(self.model.encoder, msg.payload.state_dict())

        # Deactivate gradient
        for param in self.model.head.parameters():
            param.requires_grad = False

    def fine_tune(self) -> None:
        """Fine-tune the personalized model."""

        self._load_from_cache()
        if self.hyper_params.mode == "full":
            for param in self.model.parameters():
                param.requires_grad = True
        elif self.hyper_params.mode == "body":
            for param in self.model.encoder.parameters():
                param.requires_grad = True
            for param in self.model.head.parameters():
                param.requires_grad = False
        else:  # head
            for param in self.model.encoder.parameters():
                param.requires_grad = False
            for param in self.model.head.parameters():
                param.requires_grad = True

        self.model.train()
        self.model.to(self.device)
        self.pers_optimizer, self.pers_scheduler = self._optimizer_cfg(self.model)

        for _ in range(self.hyper_params.fine_tuning_epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)
                self.pers_optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                self.pers_optimizer.step()
            self.pers_scheduler.step()

        self.model.cpu()
        clear_cuda_cache()
        self._save_to_cache()

    def finalize(self) -> None:
        self._load_from_cache()
        metrics = self.evaluate(FlukeENV().get_evaluator(), self.test_set)
        if metrics:
            self.notify(
                event="client_evaluation",
                round=-1,
                client_id=self.index,
                phase="post-fit",
                evals=metrics,
            )
        self._save_to_cache()


class FedBABUServer(Server):

    def __init__(
        self,
        model: Module,
        test_set: FastDataLoader,  # not used
        clients: Sequence[Client],
        weighted: bool = False,
    ):
        super().__init__(model=model, test_set=None, clients=clients, weighted=weighted)

    def finalize(self) -> None:

        with Progress(transient=True) as progress:
            task = progress.add_task("[cyan]Client's fine tuning", total=len(self._participants))
            clients_ft = [client for client in self.clients if client.index in self._participants]
            for client in clients_ft:
                client.fine_tune()
                client.finalize()
                progress.update(task, advance=1)

        self.notify(event="finalize", round=self.rounds + 1)


class FedBABU(PersonalizedFL):

    def get_client_class(self) -> type[Client]:
        return FedBABUClient

    def get_server_class(self) -> type[Server]:
        return FedBABUServer



"""This submodule provides logging utilities."""

import logging
import os
import sys
from typing import Any, Collection, Literal, Union

import json
import clearml
import wandb
from psutil import Process
from rich import print as rich_print
from rich.logging import RichHandler
from rich.panel import Panel
from rich.pretty import Pretty
from torch.nn import Module
from torch.utils.tensorboard import SummaryWriter

sys.path.append(".")
sys.path.append("..")

from .. import DDict  # NOQA
from ..comm import ChannelObserver, Message  # NOQA
from ..evaluation import PerformanceTracker  # NOQA
from ..utils import bytes2human, get_class_from_qualified_name  # NOQA
from . import ClientObserver, ServerObserver, get_class_from_str  # NOQA

__all__ = ["Log", "DebugLog", "TensorboardLog", "WandBLog", "ClearMLLog", "get_logger"]


class Log(ServerObserver, ChannelObserver, ClientObserver):
    """Basic logger.
    This class is used to log the performance of the global model and the communication costs during
    the federated learning process. The logging happens in the console.

    Attributes:
        tracker (PerformanceTracker): The performance tracker.
        current_round (int): The current round.
    """

    def __init__(self, **kwargs):
        self.tracker = PerformanceTracker()
        self.current_round: int = 0
        self.custom_fields: dict = {}

    def log(self, message: str) -> None:
        """Log a message.

        Args:
            message (str): The message to log.
        """
        rich_print(message)

    def add_scalar(self, key: Any, value: float, round: int) -> None:
        """Add a scalar to the logger.

        Args:
            key (Any): The key of the scalar.
            value (float): The value of the scalar.
            round (int): The round.
        """
        if round not in self.custom_fields:
            self.custom_fields[round] = {}
        self.custom_fields[round][key] = value

    def add_scalars(self, key: Any, values: dict[str, float], round: int) -> None:
        """Add scalars to the logger.

        Args:
            key (Any): The main key of the scalars.
            values (dict[str, float]): The key-value pairs of the scalars.
            round (int): The round.
        """
        if round not in self.custom_fields:
            self.custom_fields[round] = {}
        for k, v in values.items():
            self.custom_fields[round][f"{key}/{k}"] = v

    def pretty_log(self, data: Any, title: str) -> None:
        """Log a pretty-printed data.

        Args:
            data (Any): The data to log.
            title (str): The title of the data.
        """
        rich_print(Panel(Pretty(data, expand_all=True), title=title, width=100))

    def init(self, **kwargs) -> None:
        """Initialize the logger.
        The initialization is done by printing the configuration in the console.

        Args:
            **kwargs: The configuration.
        """
        if kwargs:
            rich_print(Panel(Pretty(kwargs, expand_all=True), title="Configuration", width=100))

    def start_round(self, round: int, global_model: Module) -> None:
        self.tracker.add(perf_type="comm", metrics=0, round=round)
        self.current_round = round

        if round == 1 and self.tracker.get("comm", round=0) > 0:
            rich_print(
                Panel(
                    Pretty({"comm_costs": self.tracker.get("comm", round=0)}),
                    title=f"Round: {round - 1}",
                    width=100,
                )
            )

    def end_round(self, round: int) -> None:
        stats = {
            "pre-fit": self.tracker.summary("pre-fit", round=round, force_round=False),
            "locals": self.tracker.summary("locals", round=round, force_round=False),
            "post-fit": self.tracker.summary("post-fit", round=round, force_round=False),
            "global": self.tracker.summary("global", round=round, force_round=False),
            "comm_cost": self.tracker.get("comm", round=round),
        }

        proc = Process(os.getpid())

        self.tracker.add(
            perf_type="mem",
            metrics=proc.memory_full_info().uss,
            round=round,
        )

        if self.custom_fields and round in self.custom_fields:
            stats.update(self.custom_fields[round])

        to_skip = [k for k, v in stats.items() if v is None or (isinstance(v, dict) and not v)]
        stats = {k: v for k, v in stats.items() if k not in to_skip}

        rich_print(Panel(Pretty(stats, expand_all=True), title=f"Round: {round}", width=100))
        rich_print(
            f"  Memory usage: {bytes2human(self.tracker.get('mem', round=round))} "
            + f"[{proc.memory_percent():.2f} %]"
        )

    def client_evaluation(
        self,
        round: int,
        client_id: int,
        phase: Literal["pre-fit", "post-fit"],
        evals: dict[str, float],
        **kwargs,
    ) -> None:
        if round == -1:
            round = self.current_round + 1
        self.tracker.add(perf_type=phase, metrics=evals, round=round, client_id=client_id)

    def server_evaluation(
        self,
        round: int,
        eval_type: Literal["global", "locals"],
        evals: Union[dict[str, float], dict[int, dict[str, float]]],
        **kwargs,
    ) -> None:

        self.tracker.add(perf_type=eval_type, metrics=evals, round=round)

    def message_received(self, by: Any, message: Message) -> None:
        """Update the communication costs.

        Args:
            by (Any): The sender of the message.
            message (Message): The message received.
        """
        self.tracker.add(perf_type="comm", metrics=message.size, round=self.current_round)

    def finished(self, round: int) -> None:
        stats = {
            "pre-fit": self.tracker.summary("pre-fit", round=round),
            "locals": self.tracker.summary("locals", round=round),
            "post-fit": self.tracker.summary("post-fit", round=round),
            "global": self.tracker.summary("global", round=round),
            "comm_costs": self.tracker.summary("comm", round),
        }

        to_skip = [k for k, v in stats.items() if v is None or (isinstance(v, dict) and not v)]
        stats = DDict(stats).exclude(*to_skip)

        rich_print(
            Panel(
                Pretty(stats, expand_all=True),
                title="Overall Performance",
                width=100,
            )
        )

    def interrupted(self) -> None:
        rich_print("\n[bold italic yellow]The experiment has been interrupted by the user.")

    def early_stop(self, round: int) -> None:
        return self.end_round(round)

    def track_item(self, round: int, item: str, value: Any) -> None:
        self.add_scalar(item, value, round)

    def save(self, path: str) -> None:
        """Save the logger's history to a JSON file.

        Args:
            path (str): The path to the JSON file.
        """
        json_to_save = {
            "perf_global": self.tracker["global"],
            "comm_costs": self.tracker["comm"],
            "perf_locals": self.tracker["locals"],
            "perf_prefit": self.tracker["pre-fit"],
            "perf_postfit": self.tracker["post-fit"],
            "custom_fields": self.custom_fields,
        }
        with open(path, "w") as f:
            json.dump(json_to_save, f, indent=4)

    def close(self) -> None:
        """Close the logger."""
        pass


class DebugLog(Log):
    """Debug Logger.
    This type of logger extends the basic logger by adding debug information.
    It can be seen as a more verbose version of the basic logger.

    Args:
        **kwargs: The configuration for the logger.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        logging.basicConfig(
            level=logging.DEBUG,
            format="%(message)s",
            # datefmt="[%X]",
            datefmt="[%Y-%m-%d %H:%M:%S]",
            handlers=[RichHandler(rich_tracebacks=False, show_path=False, markup=True)],
        )

        self.logger = logging.getLogger("rich")

        # Example logs
        # logger.debug("This is a [bold blue]debug[/] message.")
        # logger.info("This is a [green]info[/] message.")
        # logger.warning("This is a [yellow]warning[/] message.")
        # logger.error("This is a [red]error[/] message.")
        # logger.critical("This is a [bold red]critical[/] message.")

    def init(self, **kwargs) -> None:
        self.logger.debug("Debug logging enabled")
        super().init(**kwargs)

    def start_round(self, round: int, global_model: Module) -> None:
        self.logger.debug("Starting round %d", round)
        super().start_round(round, global_model)

    def end_round(self, round: int) -> None:
        self.logger.debug("Ending round %d", round)
        super().end_round(round)

    def selected_clients(self, round: int, clients: Collection) -> None:
        clients_idx = [client.index for client in clients]
        self.logger.debug(f"Selected {len(clients_idx)} clients for round {round}: {clients_idx}")
        super().selected_clients(round, clients)

    def server_evaluation(
        self,
        round: int,
        eval_type: Literal["global", "locals"],
        evals: Union[dict[str, float], dict[int, dict[str, float]]],
        **kwargs,
    ) -> None:
        if eval_type == "global":
            self.logger.debug(f"Global evaluation for round {round}")
        elif eval_type == "locals":
            self.logger.debug(f"Local models evaluated on server's test set for round {round}")
        super().server_evaluation(round, eval_type, evals, **kwargs)

    def finished(self, round: int) -> None:
        self.logger.debug(f"Federation ended successfully after round {round - 1}")
        return super().finished(round)

    def interrupted(self) -> None:
        self.logger.debug("Experiment interrupted by the user")
        return super().interrupted()

    def early_stop(self, round: int) -> None:
        self.logger.debug(f"Early stopping fired for round {round}")
        return super().early_stop(round)

    def start_fit(self, round: int, client_id: int, model: Module, **kwargs) -> None:
        self.logger.debug(f"Starting fit for client {client_id}")
        return super().start_fit(round, client_id, model, **kwargs)

    def end_fit(self, round: int, client_id: int, model: Module, loss: float, **kwargs) -> None:
        self.logger.debug(f"Fit for Client[{client_id}] ended with loss {loss}")
        return super().end_fit(round, client_id, model, loss, **kwargs)

    def client_evaluation(
        self,
        round: int,
        client_id: int,
        phase: Literal["pre-fit", "post-fit"],
        evals: dict[str, float],
        **kwargs,
    ) -> None:
        if round >= 0:
            self.logger.debug(f"Client[{client_id}] {phase} evaluation for round {round}")
        else:
            self.logger.debug(f"Client[{client_id}] {phase} evaluation after final round")
        self.logger.debug(evals)
        return super().client_evaluation(round, client_id, phase, evals, **kwargs)

    def message_received(self, by: Any, message: Message) -> None:
        sender = str(message.sender).split("(")[0]
        receiver = str(by).split("(")[0]
        self.logger.debug(f"Message {message.id} ({sender} -> {receiver}) received")
        return super().message_received(by, message)

    def message_sent(self, to: Any, message: Message) -> None:
        sender = str(message.sender).split("(")[0]
        receiver = str(to).split("(")[0]
        self.logger.debug(f"Message {message.id} ({sender} -> {receiver}) sent")
        return super().message_sent(to, message)

    def message_broadcasted(self, to: list[Any], message: Message) -> None:
        sender = str(message.sender).split("(")[0]
        self.logger.debug(f"Message {message.id} from {sender} broadcasted ")
        return super().message_broadcasted(to, message)


class TensorboardLog(Log):
    """TensorBoard logger.
    This class is used to log the performance of the global model and the communication costs during
    the federated learning process on TensorBoard

    See Also:
        For more information on TensorBoard, see the `official documentation
        <https://www.tensorflow.org/tensorboard>`_.

    Args:
        **config: The configuration for TensorBoard.
    """

    def __init__(self, **config):
        super().__init__(**config)
        self._config: dict = DDict(**config)
        if "log_dir" not in self._config:
            self._config["log_dir"] = "./logs"
        self._writer: SummaryWriter | None = None

    def init(self, **kwargs) -> None:
        exp_name = self._config.name
        self._config["log_dir"] = os.path.join(self._config["log_dir"], f"{exp_name}")
        self._writer = SummaryWriter(**self._config.exclude("name"))

    def add_scalar(self, key: Any, value: float, round: int) -> None:
        super().add_scalar(key, value, round)
        return self._writer.add_scalar(key, value, round)

    def add_scalars(self, key: Any, values: dict[str, float], round: int) -> None:
        super().add_scalars(key, values, round)
        return self._writer.add_scalars(key, values, round)

    def start_round(self, round: int, global_model: Module) -> None:
        super().start_round(round, global_model)
        if round == 1 and self.tracker.get("comm", round) > 0:
            self._writer.add_scalar("comm_costs", self.tracker.get("comm", round), round)
        self._writer.flush()

    def _report(self, prefix: str, evals: dict[str, float], round: int) -> None:
        for metric, value in evals.items():
            self._writer.add_scalar(f"{prefix}/{metric}", value, round)
            self._writer.flush()

    def end_round(self, round: int) -> None:
        super().end_round(round)
        global_perf = self.tracker.get("global", round)
        if global_perf is not None:
            self._report("global", global_perf, round)
        self._writer.add_scalar("comm_costs", self.tracker.get("comm", round), round)
        self._writer.flush()

        prefit_perf = self.tracker.summary("pre-fit", round=round, include_round=True)
        if prefit_perf:
            self._report("pre-fit", prefit_perf, prefit_perf["round"])

        postfit_perf = self.tracker.summary("post-fit", round=round, include_round=True)
        if postfit_perf:
            self._report("post-fit", postfit_perf, postfit_perf["round"])

        locals_perf = self.tracker.summary("locals", round=round, include_round=True)
        if locals_perf:
            self._report("locals", locals_perf, locals_perf["round"])

        self._writer.flush()

    def finished(self, round: int) -> None:
        super().finished(round)
        prefit_perf = self.tracker.summary("pre-fit", round=round, include_round=True)
        if prefit_perf and round == prefit_perf["round"]:
            self._report("pre-fit", prefit_perf, round)

        locals_perf = self.tracker.summary("locals", round=round, include_round=True)
        if locals_perf and round == locals_perf["round"]:
            self._report("locals", locals_perf, round)

    def close(self) -> None:
        self._writer.flush()
        self._writer.close()


class WandBLog(Log):
    """Weights and Biases logger.
    This class is used to log the performance of the global model and the communication costs during
    the federated learning process on Weights and Biases.

    See Also:
        For more information on Weights and Biases, see the `Weights and Biases documentation
        <https://docs.wandb.ai/>`_.

    Args:
        **config: The configuration for Weights and Biases.
    """

    def __init__(self, **config):
        super().__init__(**config)
        self.config = config
        self.run = None

    def init(self, **kwargs) -> None:
        super().init(**kwargs)
        self.config["config"] = kwargs
        self.run = wandb.init(**self.config)

    def add_scalar(self, key: Any, value: float, round: int) -> None:
        super().add_scalar(key, value, round)
        return self.run.log({key: value}, step=round)

    def add_scalars(self, key: Any, values: dict[str, float], round: int) -> None:
        super().add_scalars(key, values, round)
        return self.run.log({f"{key}/{k}": v for k, v in values.items()}, step=round)

    def start_round(self, round: int, global_model: Module) -> None:
        super().start_round(round, global_model)
        if round == 1 and self.tracker.get("comm", round) > 0:
            self.run.log({"comm_costs": self.tracker.get("comm", round)})

    def end_round(self, round: int) -> None:
        super().end_round(round)
        global_perf = self.tracker.get("global", round)
        if global_perf is not None:
            self.run.log({"global": global_perf}, step=round)
        self.run.log({"comm_cost": self.tracker.get("comm", round)}, step=round)

        prefit_perf = self.tracker.summary("pre-fit", round=round, include_round=True)
        if prefit_perf:
            self.run.log({"pre-fit": prefit_perf}, step=prefit_perf["round"])

        postfit_perf = self.tracker.summary("post-fit", round=round, include_round=True)
        if postfit_perf:
            self.run.log({"post-fit": postfit_perf}, step=postfit_perf["round"])

        locals_perf = self.tracker.summary("locals", round=round, include_round=True)
        if locals_perf:
            self.run.log({"locals": locals_perf}, step=locals_perf["round"])

    def finished(self, round: int) -> None:
        super().finished(round)

        server_last_round = max(self.tracker["global"])
        self.run.log(
            {"global": self.tracker.get("global", server_last_round)}, step=server_last_round
        )

        postfit_eval_summary = self.tracker.summary("post-fit", round=round, include_round=True)
        if postfit_eval_summary:
            last_round = postfit_eval_summary["round"]
            # avoid warning
            if server_last_round <= last_round:
                self.run.log({"post-fit": postfit_eval_summary}, step=last_round)

        prefit_eval_summary = self.tracker.summary("pre-fit", round=round, include_round=True)
        if prefit_eval_summary:
            last_round = prefit_eval_summary["round"]
            # avoid warning
            if server_last_round <= last_round:
                self.run.log({"pre-fit": prefit_eval_summary}, step=last_round)

        locals_eval_summary = self.tracker.summary("locals", round=round, include_round=True)
        if locals_eval_summary:
            last_round = locals_eval_summary["round"]
            # avoid warning
            if server_last_round <= last_round:
                self.run.log({"locals": locals_eval_summary}, step=last_round)

    def close(self) -> None:
        self.run.finish()

    def save(self, path: str) -> None:
        super().save(path)


class ClearMLLog(TensorboardLog):
    """ClearML logger.
    This class is used to log the performance of the global model and the communication costs during
    the federated learning process on ClearML.

    Note:
        The ClearML logger takes advantage of the TensorBoard logger, thus the logging also happens
        on TensorBoard. The logging folder is "./runs/{experiment_name}_{timestamp}".

    See Also:
        For more information on ClearML, see the `official documentation
        <https://clear.ml/docs/latest/docs/>`_.

    Args:
        **config: The configuration for ClearML.
    """

    def __init__(self, **config):
        super().__init__(name=config["name"])
        self.config: DDict = DDict(**config)
        self.task: clearml.task.Task | None = None

    def init(self, **kwargs) -> None:
        super().init(**kwargs)
        # imported here to avoid issues with requests
        from clearml import Task

        self.task = Task.init(task_name=self.config.name, **self.config.exclude("name"))
        self.task.connect(kwargs)

    def close(self) -> None:
        super().close()
        if self.task is not None:
            self.task.close()
            self.task = None


def get_logger(lname: str, **kwargs) -> Log:
    """Get a logger from its name.
    This function is used to get a logger from its name. It is used to dynamically import loggers.
    The supported loggers are the ones defined in the ``fluke.utils.log`` module, but it can handle
    any logger defined by the user.

    Note:
        To use a custom logger, it must be defined in a module and the full model name must be
        provided in the configuration file. For example, if the logger ``MyLogger`` is defined in
        the module ``my_module`` (i.e., a file called ``my_module.py``), the logger name must be
        ``my_module.MyLogger``. The other logger's parameters must be passed as in the following
        example:

        .. code-block:: yaml

            logger:
                name: my_module.MyLogger
                param1: value1
                param2: value2
                ...


    Args:
        lname (str): The name of the logger.
        **kwargs: The keyword arguments to pass to the logger's constructor.

    Returns:
        Log | DebugLog | WandBLog | ClearMLLog | TensorboardLog: The logger.
    """
    if "." in lname and not lname.startswith("fluke.utils.log"):
        return get_class_from_qualified_name(lname)(**kwargs)
    return get_class_from_str("fluke.utils.log", lname)(**kwargs)



"""Implementation of the [FedSGD17]_ algorithm.

References:
    .. [FedSGD17] H. Brendan McMahan, Eider Moore, Daniel Ramage, Seth Hampson, and Blaise Aguera
       y Arcas. Communication-efficient learning of deep networks from decentralized data.
       In AISTATS (2017). URL: https://arxiv.org/abs/1602.05629
"""

import sys

import torch

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA

__all__ = ["FedSGDClient", "FedSGD"]


class FedSGDClient(Client):
    def __init__(
        self,
        index: int,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: torch.nn.Module,
        local_epochs: int = 3,
        fine_tuning_epochs: int = 0,
        **kwargs,
    ):
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            **kwargs,
        )
        self.train_set.single_batch = True
        self.train_set.shuffle = True
        self.hyper_params.local_epochs = 1


class FedSGD(CentralizedFL):

    def get_client_class(self) -> type[Client]:
        return FedSGDClient



"""Implementation of the FedALA [FedALA23]_ algorithm.

References:
    .. [FedALA23] Jianqing Zhang, Yang Hua, Hao Wang, Tao Song, Zhengui Xue, Ruhui Ma, Haibing Guan.
       FedALA: Adaptive Local Aggregation for Personalized Federated Learning
       In AAAI (2023). URL: https://arxiv.org/pdf/2212.01197v4

"""

import sys
from copy import deepcopy

import numpy as np
import torch
from torch import nn

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderHeadNet  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA
from . import CentralizedFL  # NOQA

__all__ = ["FedALAClient", "FedALA"]


class FedALAClient(Client):

    def __init__(
        self,
        index: int,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: nn.Module,
        local_epochs: int = 3,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        ala_sample_size: float = 0.8,
        eta: float = 1.0,
        conergence_threshold: float = 0.001,
        loss_window_size: int = 10,
        **kwargs,
    ):
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.hyper_params.update(
            ala_sample_size=ala_sample_size,
            eta=eta,
            conergence_threshold=conergence_threshold,
            loss_window_size=loss_window_size,
        )
        self.weights = None
        self.start_phase = True

    def adaptive_local_aggregation(self, server_model: EncoderHeadNet) -> None:
        # keep the server encoder weights
        safe_load_state_dict(self.model.encoder, server_model.encoder.state_dict())

        random_sample_loader = FastDataLoader(
            *self.train_set.tensors,
            num_labels=self.train_set.num_labels,
            batch_size=self.train_set.batch_size,
            shuffle=True,
            percentage=self.hyper_params.ala_sample_size,
        )

        temp_model = deepcopy(self.model)

        self.model.to(self.device)
        temp_model.to(self.device)
        server_model.to(self.device)

        # frozen the encoder weights
        for param in temp_model.encoder.parameters():
            param.requires_grad = False

        optimizer = torch.optim.SGD(temp_model.head.parameters(), lr=0)

        if self.weights is None:
            self.weights = [
                torch.ones_like(p.data).to(self.device) for p in self.model.head.parameters()
            ]

        # initialize the temp model
        for param_t, param, param_g, weight in zip(
            temp_model.head.parameters(),
            self.model.head.parameters(),
            server_model.head.parameters(),
            self.weights,
        ):
            param_t.data = param + (param_g - param) * weight

        converged = False
        losses = []
        while not converged and len(losses) < 100:

            for X, y in random_sample_loader:
                X, y = X.to(self.device), y.to(self.device)
                optimizer.zero_grad()
                y_hat = temp_model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()
                # optimizer.step()

                losses.append(loss.item())

                # update weight in this batch
                for param_t, param, param_g, weight in zip(
                    temp_model.head.parameters(),
                    self.model.head.parameters(),
                    server_model.head.parameters(),
                    self.weights,
                ):
                    weight.data = torch.clamp(
                        weight - self.hyper_params.eta * (param_t.grad * (param_g - param)), 0, 1
                    )

                # update temp local model in this batch
                for param_t, param, param_g, weight in zip(
                    temp_model.head.parameters(),
                    self.model.head.parameters(),
                    server_model.head.parameters(),
                    self.weights,
                ):
                    param_t.data = param + (param_g - param) * weight

            # only train one epoch in the subsequent iterations
            if not self.start_phase:
                break

            # train the weight until convergence
            if len(losses) > self.hyper_params.loss_window_size:
                loss_std = np.std(losses[-self.hyper_params.loss_window_size :])
                if loss_std < self.hyper_params.conergence_threshold:
                    converged = True

        self.start_phase = False
        self.model.cpu()
        temp_model.cpu()
        server_model.cpu()
        safe_load_state_dict(self.model.head, temp_model.head.state_dict())

    def receive_model(self) -> None:
        server_model = self.channel.receive(self.index, "server", msg_type="model").payload
        if self.model is None:
            self.model = server_model
        else:
            self.adaptive_local_aggregation(server_model)


class FedALA(CentralizedFL):

    def get_client_class(self) -> type[Client]:
        return FedALAClient



import sys
from typing import Any

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..client import Client  # NOQA
from ..evaluation import Evaluator  # NOQA
from ..utils.model import ModOpt  # NOQA
from ..utils.model import optimizer_to, safe_load_state_dict  # NOQA
from .utils import ModelBuilder  # NOQA

__all__ = ["ParallelClient"]


class ParallelClient(Client):
    """A client that can be used in a parallel setting."""

    def __init__(self, builder: ModelBuilder, *args: Any, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._builder = builder

    def __getstate__(self) -> dict[str, Any]:
        """Get the state of the client as a dictionary. This method is used to serialize the client
        when using multiprocessing.

        Returns:
            dict: The state of the client.
        """
        state = self.state_dict()
        state["optimizer_cfg"] = self._optimizer_cfg
        state["device"] = str(self.device)
        state["train_set"] = self.train_set
        state["test_set"] = self.test_set
        state["hyper_params"] = self.hyper_params
        state["builder"] = self._builder
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Set the state of the client from a dictionary. This method is used to deserialize the
        client when using multiprocessing.

        Args:
            state (dict): The state of the client.
        """
        self._modopt = ModOpt()
        self._builder = state["builder"]
        if self.model is None:
            self.model = self._builder.build()
        self._optimizer_cfg = state["optimizer_cfg"]
        if state["modopt"]["model"] is not None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)
            self._modopt.load_state_dict(state["modopt"])

        self._index = state["index"]
        self._last_round = state["last_round"]
        self.device = torch.device(state["device"])
        self.train_set = state["train_set"]
        self.test_set = state["test_set"]
        self.hyper_params = state["hyper_params"]

    def local_update(
        self,
        current_round: int,
        current_model_sd: dict,
        device: str,
        prefit: bool,
        postfit: bool,
        evaluator: Evaluator,
    ) -> tuple[Module, float, int]:
        self.device = device

        self._last_round = current_round
        safe_load_state_dict(self.model, current_model_sd)

        eval_results = {}
        if evaluator is not None and prefit:
            metrics = self.evaluate(evaluator, self.test_set)
            if metrics:
                eval_results["pre-fit"] = metrics

        loss = self.fit()
        optimizer_to(self.optimizer, "cpu")

        if evaluator is not None and postfit:
            metrics = self.evaluate(evaluator, self.test_set)
            if metrics:
                eval_results["post-fit"] = metrics

        return self._modopt, loss, self.index, eval_results

    def send_model(self):
        raise NotImplementedError(
            "ParallelClient does not support sending model to server using the channel."
        )

    def receive_model(self, model: Module) -> None:
        raise NotImplementedError(
            "ParallelClient does not support receiving model from server using the channel."
        )



"""Implementation of the [SuPerFed22]_ algorithm.

References:
    .. [SuPerFed22] Seok-Ju Hahn, Minwoo Jeong, and Junghye Lee. Connecting Low-Loss Subspace
       for Personalized Federated Learning. In KDD (2022). URL: https://arxiv.org/abs/2109.07628v3
"""

import sys
from copy import deepcopy

import numpy as np
from torch.nn.modules import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import PersonalizedFL  # NOQA
from ..client import PFLClient  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..utils import clear_cuda_cache  # NOQA
from ..utils.model import get_global_model_dict  # NOQA
from ..utils.model import mix_networks, safe_load_state_dict, set_lambda_model  # NOQA

__all__ = ["SuPerFedClient", "SuPerFed"]


class SuPerFedClient(PFLClient):

    def __init__(
        self,
        index: int,
        model: Module,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: Module,
        local_epochs: int = 3,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        mode: str = "global",
        start_mix: int = 10,
        mu: float = 0.1,
        nu: float = 0.1,
        **kwargs,
    ):
        assert mode in ["mm", "lm"]

        super().__init__(
            index=index,
            model=mix_networks(model, deepcopy(model), 0),
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.hyper_params.update(mode=mode, start_mix=start_mix, mu=mu, nu=nu)

    def receive_model(self) -> None:
        msg = self.channel.receive(self.index, "server", msg_type="model")
        if self.model is None:
            self.model = msg.payload
        else:
            safe_load_state_dict(self.model, msg.payload.state_dict())

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (
            override_local_epochs if override_local_epochs > 0 else self.hyper_params.local_epochs
        )

        if self._last_round >= self.hyper_params.start_mix:
            local_dict = {
                k: v for k, v in self.personalized_model.state_dict().items() if "_local" in k
            }
        else:
            local_dict = {k + "_local": v for k, v in self.model.state_dict().items()}
        self.personalized_model.load_state_dict({**self.model.state_dict(), **local_dict})

        self.personalized_model.train()
        self.personalized_model.to(self.device)

        if self.pers_optimizer is None:
            self.pers_optimizer, self.pers_scheduler = self._optimizer_cfg(self.personalized_model)

        if self.hyper_params.mu > 0:
            prev_global_model = self.model.to(self.device)
            for param in prev_global_model.parameters():
                param.requires_grad = False

        running_loss = 0.0
        for _ in range(epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                X, y = X.to(self.device), y.to(self.device)

                if self._last_round >= self.hyper_params.start_mix:
                    if self.hyper_params.mode == "mm":
                        set_lambda_model(self.personalized_model, np.random.uniform(0.0, 1.0))
                    elif self.hyper_params.mode == "lm":
                        set_lambda_model(self.personalized_model, 0.0, layerwise=True)

                self.pers_optimizer.zero_grad()
                y_hat = self.personalized_model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)

                if self.hyper_params.mu > 0:
                    prox_term = 0.0
                    for name, param in self.personalized_model.named_parameters():
                        if "_local" not in name:
                            continue
                        internal_param = self.personalized_model.get_parameter(
                            name.replace("_local", "")
                        )
                        prev_param = prev_global_model.get_parameter(name.replace("_local", ""))
                        prox_term += (internal_param - prev_param).norm(2)
                    loss += self.hyper_params.mu * prox_term

                if self._last_round >= self.hyper_params.start_mix:
                    numerator, norm_1, norm_2 = 0, 0, 0
                    for name, param_l in self.personalized_model.named_parameters():
                        if "_local" not in name:
                            continue
                        param_g = self.personalized_model.get_parameter(name.replace("_local", ""))
                        numerator += (param_g * param_l).add(1e-6).sum()
                        norm_1 += param_g.pow(2).sum()
                        norm_2 += param_l.pow(2).sum()
                    cos_sim = numerator.pow(2).div(norm_1 * norm_2)
                    loss += self.hyper_params.nu * cos_sim

                loss.backward()
                self._clip_grads(self.personalized_model)
                self.pers_optimizer.step()
                running_loss += loss.item()

            self.pers_scheduler.step()

        running_loss /= epochs * len(self.train_set)
        self.personalized_model.cpu()
        clear_cuda_cache()

        set_lambda_model(self.personalized_model, 1.0)
        self.model.load_state_dict(get_global_model_dict(self.personalized_model))

        return running_loss


class SuPerFed(PersonalizedFL):

    def get_client_class(self) -> type[PFLClient]:
        return SuPerFedClient



"""Implementation of the [FedPer19]_ algorithm.

References:
    .. [FedPer19] Manoj Ghuhan Arivazhagan, Vinay Aggarwal, Aaditya Kumar Singh, and
       Sunav Choudhary. Federated learning with personalization layers.
       In arXiv (2019). URL:https://arxiv.org/abs/1912.00818
"""

import sys
from copy import deepcopy
from typing import Sequence

import torch

sys.path.append(".")
sys.path.append("..")

from ..algorithms import PersonalizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..nets import EncoderGlobalHeadLocalNet, EncoderHeadNet  # NOQA
from ..server import Server  # NOQA
from ..utils.model import safe_load_state_dict  # NOQA


# https://arxiv.org/abs/1912.00818
class FedPerClient(Client):

    def __init__(
        self,
        index: int,
        model: EncoderHeadNet,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: torch.nn.Module,
        local_epochs: int = 3,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        **kwargs,
    ):
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.model = EncoderGlobalHeadLocalNet(model)
        self._save_to_cache()

    def send_model(self) -> None:
        self.channel.send(
            Message(deepcopy(self.model.get_global()), "model", self.index, inmemory=True),
            "server",
        )

    def receive_model(self) -> None:
        msg = self.channel.receive(self.index, "server", msg_type="model")
        safe_load_state_dict(self.model.get_global(), msg.payload.state_dict())


class FedPerServer(Server):

    def __init__(
        self,
        model: torch.nn.Module,
        test_set: FastDataLoader,  # not used
        clients: Sequence[Client],
        weighted: bool = False,
    ):
        super().__init__(model=model, test_set=None, clients=clients, weighted=weighted)


class FedPer(PersonalizedFL):

    def get_client_class(self) -> type[Client]:
        return FedPerClient

    def get_server_class(self) -> type[Server]:
        return FedPerServer



"""Implementation of the [SCAFFOLD20]_ algorithm.

References:
    .. [SCAFFOLD20] Sai Praneeth Karimireddy, Satyen Kale, Mehryar Mohri, Sashank J. Reddi,
       Sebastian U. Stich, Ananda Theertha Suresh. SCAFFOLD: Stochastic Controlled Averaging for
       Federated Learning. In ICML (2020). URL: https://arxiv.org/abs/1910.06378
"""

import sys
from collections import OrderedDict
from copy import deepcopy
from typing import Collection, Sequence

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from .. import FlukeENV  # NOQA
from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..comm import Message  # NOQA
from ..config import OptimizerConfigurator  # NOQA
from ..data import FastDataLoader  # NOQA
from ..server import Server  # NOQA
from ..utils import clear_cuda_cache  # NOQA
from ..utils.model import safe_load_state_dict, state_dict_zero_like, unwrap  # NOQA

__all__ = ["SCAFFOLDClient", "SCAFFOLDServer", "SCAFFOLD"]


class SCAFFOLDClient(Client):
    def __init__(
        self,
        index: int,
        train_set: FastDataLoader,
        test_set: FastDataLoader,
        optimizer_cfg: OptimizerConfigurator,
        loss_fn: torch.nn.Module,
        local_epochs: int = 3,
        fine_tuning_epochs: int = 0,
        clipping: float = 0,
        **kwargs,
    ):
        super().__init__(
            index=index,
            train_set=train_set,
            test_set=test_set,
            optimizer_cfg=optimizer_cfg,
            loss_fn=loss_fn,
            local_epochs=local_epochs,
            fine_tuning_epochs=fine_tuning_epochs,
            clipping=clipping,
            **kwargs,
        )
        self.control: OrderedDict | None = None
        self.delta_control: OrderedDict | None = None
        self.server_control: OrderedDict | None = None
        self._attr_to_cache.extend(["control", "delta_control", "server_control"])

    def receive_model(self) -> None:
        model = self.channel.receive(self.index, "server", msg_type="model").payload
        self.server_control = self.channel.receive(self.index, "server", msg_type="control").payload
        if self.model is None:
            self.model = model
            if self.control is None:
                self.control = state_dict_zero_like(model.state_dict())
        else:
            safe_load_state_dict(self.model, model.state_dict())
        self.server_model = deepcopy(model.state_dict())

    def fit(self, override_local_epochs: int = 0) -> float:
        epochs: int = (
            override_local_epochs if override_local_epochs > 0 else self.hyper_params.local_epochs
        )
        self.model.to(self.device)
        self.model.train()

        if self.optimizer is None:
            self.optimizer, self.scheduler = self._optimizer_cfg(self.model)

        K = 0
        running_loss = 0.0
        for _ in range(epochs):
            loss = None
            for _, (X, y) in enumerate(self.train_set):
                K += 1
                X, y = X.to(self.device), y.to(self.device)
                self.optimizer.zero_grad()
                y_hat = self.model(X)
                loss = self.hyper_params.loss_fn(y_hat, y)
                loss.backward()

                for n, p in unwrap(self.model).named_parameters():
                    p.grad.data = p.grad.data + (
                        self.server_control[n].to(self.device) - self.control[n].to(self.device)
                    )

                self._clip_grads(self.model)
                self.optimizer.step()
                running_loss += loss.item()
            self.scheduler.step()

        self.model.cpu()
        # This happens on CPU!

        with torch.no_grad():
            c_plus = state_dict_zero_like(self.control)
            c_delta = state_dict_zero_like(self.control)
            model_params = unwrap(self.model).state_dict()
            for key in model_params:
                c_plus[key] = (
                    self.control[key]
                    - self.server_control[key]
                    + (self.server_model[key] - model_params[key])
                    / (K * self.scheduler.get_last_lr()[0])
                )

            for key in model_params:
                c_delta[key] = c_plus[key] - self.control[key]

            self.control = deepcopy(c_plus)
            self.delta_control = c_delta

        running_loss /= epochs * len(self.train_set)
        clear_cuda_cache()
        return running_loss

    def send_model(self) -> None:
        self.channel.send(Message(self.model, "model", self.index, inmemory=True), "server")
        self.channel.send(
            Message(self.delta_control, "control", self.index, inmemory=True), "server"
        )


class SCAFFOLDServer(Server):
    def __init__(
        self,
        model: Module,
        test_set: FastDataLoader,
        clients: Sequence[Client],
        weighted: bool = True,
        global_step: float = 1.0,
        **kwargs,
    ):
        super().__init__(
            model=model, test_set=test_set, clients=clients, weighted=weighted, **kwargs
        )
        self.device = FlukeENV().get_device()
        self.control = state_dict_zero_like(self.model.state_dict())
        self.hyper_params.update(global_step=global_step)

    def broadcast_model(self, eligible: Sequence[Client]) -> None:
        self.channel.broadcast(Message(self.model, "model", "server"), [c.index for c in eligible])
        self.channel.broadcast(
            Message(self.control, "control", "server"), [c.index for c in eligible]
        )

    def _get_client_weights(self, eligible: Sequence[Client]) -> list[float]:
        weights = super()._get_client_weights(eligible)
        if self.hyper_params.global_step != 1:
            for c in range(len(eligible)):
                weights[c] = self.hyper_params.global_step * weights[c]
        return weights

    @torch.no_grad()
    def aggregate(self, eligible: Sequence[Client], client_models: Collection[Module]) -> None:
        self.model.to(self.device)

        total_delta = state_dict_zero_like(self.model.state_dict())
        for client in eligible:
            c_delta_params = self.channel.receive("server", client.index, "control").payload
            for key in total_delta:
                total_delta[key] = total_delta[key] + c_delta_params[key].to(self.device)

        for key in total_delta:
            total_delta[key] = total_delta[key] / self.n_clients
            self.control[key] = self.control[key] + total_delta[key].to("cpu")

        self.model.cpu()
        return super().aggregate(eligible, client_models)


class SCAFFOLD(CentralizedFL):

    def get_client_class(self) -> type[Client]:
        return SCAFFOLDClient

    def get_server_class(self) -> type[Server]:
        return SCAFFOLDServer



import gc
import shutil
import sys
import tempfile
from typing import Any

import numpy as np
from torch.nn import CrossEntropyLoss, Module
from torch.optim import SGD

sys.path.append(".")
sys.path.append("..")

from fluke import DDict, FlukeENV  # NOQA
from fluke.algorithms import CentralizedFL, PersonalizedFL  # NOQA
from fluke.client import Client  # NOQA
from fluke.comm import ChannelObserver, Message  # NOQA
from fluke.config import Configuration  # NOQA
from fluke.data import DataSplitter  # NOQA
from fluke.data.datasets import Datasets  # NOQA
from fluke.evaluation import ClassificationEval  # NOQA
from fluke.nets import MNIST_2NN  # NOQA
from fluke.server import Server  # NOQA
from fluke.utils import ClientObserver, ServerObserver, get_class_from_qualified_name  # NOQA
from fluke.utils.log import Log  # NOQA

FlukeENV().set_evaluator(ClassificationEval(1, 10))
FlukeENV().set_eval_cfg(post_fit=True, pre_fit=True)
FlukeENV().set_save_options(path="tests/tmp/tmp", save_every=1, global_only=True)


def test_centralized_fl():
    hparams = DDict(
        # model="fluke.nets.MNIST_2NN",
        model=MNIST_2NN(),
        client=DDict(
            batch_size=32,
            local_epochs=1,
            loss="CrossEntropyLoss",
            optimizer=DDict(name="SGD", lr=0.1, momentum=0.9),
            scheduler=DDict(step_size=1, gamma=0.1),
        ),
        server=DDict(weighted=True),
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist, client_split=0.1)
    fl = CentralizedFL(2, splitter, hparams)

    assert fl.n_clients == 2
    assert fl.hyper_params.match(hparams)
    assert len(fl.clients) == 2
    assert isinstance(fl.clients[0], Client)
    assert isinstance(fl.server, Server)
    assert isinstance(fl.server.model, MNIST_2NN)
    assert fl.clients[0].test_set is not None
    assert fl.clients[1].test_set is not None
    assert isinstance(fl.clients[0].hyper_params.loss_fn, CrossEntropyLoss)

    hparams = DDict(
        model="fluke.nets.MNIST_2NN",
        client=DDict(
            batch_size=32,
            local_epochs=1,
            loss="CrossEntropyLoss",
            optimizer=DDict(lr=0.1, momentum=0.9),
            scheduler=DDict(step_size=1, gamma=0.1),
        ),
        server=DDict(weighted=True),
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist, client_split=0.1)
    fl = CentralizedFL(2, splitter, hparams)

    assert isinstance(fl.server.model, MNIST_2NN)
    assert isinstance(fl.clients[0].hyper_params.loss_fn, CrossEntropyLoss)

    assert fl.get_optimizer_class() == SGD
    assert fl.get_client_class() == Client
    assert fl.get_server_class() == Server

    class Observer(ClientObserver, ServerObserver, ChannelObserver):
        def __init__(self):
            super().__init__()
            self.called_start = False
            self.called_end = False
            self.called_selected = False
            self.called_error = False
            self.called_finished = False
            self.called_client_eval = False
            self.called_server_eval = False
            self.called_start_fit = False
            self.called_end_fit = False

        def start_round(self, round, global_model):
            assert round == 1
            assert global_model is not None
            self.called_start = True

        def end_round(self, round):
            assert round == 1
            self.called_end = True

        def server_evaluation(self, round, eval_type, evals) -> None:
            assert round == 1
            assert eval_type == "global"
            assert "accuracy" in evals
            self.called_server_eval = True

        def client_evaluation(self, round, client_id, phase, evals, **kwargs):
            # assert round == 1 or (round == -1 and phase == "pre-fit")
            assert phase == "post-fit" or phase == "pre-fit"
            assert client_id == 0 or client_id == 1
            self.called_client_eval = True

        def selected_clients(self, round, clients):
            assert len(clients) == 1
            assert round == 1
            self.called_selected = True

        def finished(self, round):
            assert round == 2
            self.called_finished = True

        def start_fit(self, round: int, client_id: int, model: Module, **kwargs):
            assert round == 1
            assert client_id == 0 or client_id == 1
            self.called_start_fit = True

        def end_fit(self, round: int, client_id: int, model: Module, loss: float, **kwargs):
            assert round == 1
            assert client_id == 0 or client_id == 1
            assert loss >= 0.0
            self.called_end_fit = True

        def message_received(self, by: Any, message: Message):
            pass

    obs = Observer()
    fl.set_callbacks(obs)

    assert fl.server._observers == [fl, obs]

    strfl = (
        f"CentralizedFL[{fl.id}](model=fluke.nets.MNIST_2NN(),Client[0-1](optim=OptCfg(SGD,lr=0.1,"
        + "momentum=0.9,StepLR(step_size=1,gamma=0.1)),batch_size=32,loss_fn=CrossEntropyLoss(),"
        + "local_epochs=1,fine_tuning_epochs=0,clipping=0),Server(weighted=True,lr=1.0))"
    )

    assert str(fl).replace(" ", "").replace("\n", "").replace("\t", "") == strfl
    assert fl.__repr__().replace(" ", "").replace("\n", "").replace("\t", "") == strfl

    fl.run(1, 0.5)
    assert obs.called_start
    assert obs.called_end
    assert obs.called_selected
    assert obs.called_finished
    assert obs.called_start_fit
    assert obs.called_end_fit
    assert obs.called_client_eval
    assert obs.called_server_eval

    with tempfile.TemporaryDirectory() as tmpdirname:
        temppath = fl.save(tmpdirname)
        fl2 = CentralizedFL(2, splitter, hparams)
        fl2.load(temppath)

        assert fl2.server.rounds == fl.server.rounds

    shutil.rmtree(f"tests/tmp/tmp_{fl.id}")

    FlukeENV().set_inmemory(True)
    hparams = DDict(
        # model="fluke.nets.MNIST_2NN",
        model=MNIST_2NN(),
        client=DDict(
            batch_size=32,
            local_epochs=1,
            model=MNIST_2NN(),
            loss=CrossEntropyLoss,
            optimizer=DDict(name=SGD, lr=0.1, momentum=0.9),
            scheduler=DDict(step_size=1, gamma=0.1),
        ),
        server=DDict(weighted=True),
    )
    mnist = Datasets.MNIST("../data")
    splitter = DataSplitter(mnist)
    fl = PersonalizedFL(2, splitter, hparams)

    assert fl.n_clients == 2
    assert fl.hyper_params == hparams
    assert len(fl.clients) == 2
    assert isinstance(fl.clients[0], Client)
    assert isinstance(fl.server, Server)
    assert isinstance(fl.server.model, MNIST_2NN)
    # assert isinstance(fl.clients[0].model, MNIST_2NN)
    assert fl.clients[0].test_set is None
    assert isinstance(fl.clients[0].hyper_params.loss_fn, CrossEntropyLoss)

    fl.run(1, 0.5)
    with tempfile.TemporaryDirectory() as tmpdirname:
        temppath = fl.save(tmpdirname)
        fl2 = PersonalizedFL(2, splitter, hparams)
        fl2.load(temppath)

        assert fl2.server.rounds == fl.server.rounds

    shutil.rmtree(f"tests/tmp/tmp_{fl.id}")


def get_splitter(cfg):
    dataset = Datasets.get(**cfg.data.dataset)
    splitter = DataSplitter(
        dataset=dataset,
        distribution=cfg.data.distribution.name,
        dist_args=cfg.data.distribution.exclude("name"),
        **cfg.data.exclude("dataset", "distribution"),
    )
    return splitter


SPLITTER = None


def _test_algo(exp_config, alg_config, oncpu=True, tol=1e-5):
    accs = []
    cfg = Configuration(exp_config, alg_config)
    if oncpu:
        cfg.exp.device = "cpu"
    else:
        # if torch.cuda.is_available():
        #     cfg.exp.device = "cuda"
        # elif torch.backends.mps.is_available():
        #     cfg.exp.device = "mps"
        # else:
        return None, None
    FlukeENV().configure(cfg)
    gc.collect()
    global SPLITTER
    if SPLITTER is None:
        dataset = Datasets.get(**cfg.data.dataset)
        SPLITTER = DataSplitter(
            dataset=dataset,
            distribution=cfg.data.distribution.name,
            dist_args=cfg.data.distribution.exclude("name"),
            **cfg.data.exclude("dataset", "distribution"),
        )
    fl_algo_class = get_class_from_qualified_name(cfg.method.name)

    for mem in [False, True]:
        FlukeENV().configure(cfg)
        cfg.exp.inmemory = mem
        FlukeENV().set_inmemory(mem)
        algo = fl_algo_class(cfg.protocol.n_clients, SPLITTER, cfg.method.hyperparameters)

        log = Log()
        log.init(**cfg)
        algo.set_callbacks(log)
        algo.run(cfg.protocol.n_rounds, cfg.protocol.eligible_perc)
        FlukeENV().close_cache()
        shutil.rmtree(f"tests/tmp/tmp_{algo.id}")
        del algo
        if log.tracker["global"] and log.tracker.get("global", cfg.protocol.n_rounds):
            accs.append(log.tracker.get("global", cfg.protocol.n_rounds)["accuracy"])
        else:
            accs.append(log.tracker.summary("post-fit", cfg.protocol.n_rounds)["accuracy"])
    assert np.allclose(accs[0], accs[1], atol=tol)
    return None, log


def test_apfl():
    apfl, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/apfl.yaml")
    # apfl, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/apfl.yaml", oncpu=False)


def test_ccvr():
    ccvr, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/ccvr.yaml")
    # ccvr, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/ccvr.yaml", oncpu=False)


def test_ditto():
    ditto, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/ditto.yaml")
    # ditto, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/ditto.yaml", oncpu=False)


def test_fat():
    fat, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fat.yaml")


def test_fedala():
    fedala, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedala.yaml")
    # fedala, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedala.yaml", oncpu=False)


def test_fedamp():
    fedamp, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedamp.yaml")
    # fedamp, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedamp.yaml", oncpu=False)


def test_fedavg():
    fedavg, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedavg.yaml")
    # fedavg, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedavg.yaml", oncpu=False)


def test_fedavgm():
    fedavgm, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedavgm.yaml")
    # fedavgm, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedavgm.yaml", oncpu=False)


def test_fedaws():
    fedaws, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedaws.yaml")
    # fedaws, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedaws.yaml", oncpu=False)


def test_fedbabu():
    fedbabu, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedbabu.yaml")
    fedbabu, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedbabu_head.yaml")
    fedbabu, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedbabu_body.yaml")
    # fedbabu, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedbabu.yaml", oncpu=False)
    # fedbabu, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedbabu_head.yaml", oncpu=False)
    # fedbabu, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedbabu_body.yaml", oncpu=False)


def test_fedbn():
    fedbn, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedbn.yaml")
    # fedbn, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fedbn.yaml", oncpu=False)


def test_feddyn():
    feddyn, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/feddyn.yaml")
    # feddyn, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/feddyn.yaml", oncpu=False)


def test_fedexp():
    fedexp, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedexp.yaml")
    # fedexp, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedexp.yaml", oncpu=False)


def test_fedhp():
    fedhp, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedhp.yaml")
    # fedhp, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fedhp.yaml", oncpu=False)


def test_fedlc():
    fedlc, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedlc.yaml")
    # fedlc, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fedlc.yaml", oncpu=False)


def test_fedld():
    fedlc, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedld.yaml")
    # fedlc, log = _test_algo("./tests/configs/exp.yaml",
    #


def test_fednh():
    fednh, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fednh.yaml")
    # fednh, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fednh.yaml", oncpu=False)


def test_fednova():
    fednova, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fednova.yaml")
    # fednova, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fednova.yaml", oncpu=False)


def test_fedopt():
    fedopt, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedadam.yaml")
    fedopt, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedadagrad.yaml")
    fedopt, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedyogi.yaml")
    # fedopt, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedadam.yaml", oncpu=False)
    # fedopt, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedadagrad.yaml", oncpu=False)
    # fedopt, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedyogi.yaml", oncpu=False)


def test_fedproto():
    fedproto, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedproto.yaml")
    # fedproto, log = _test_algo("./tests/configs/exp.yaml",
    #                            "./tests/configs/alg/fedproto.yaml", oncpu=False)


def test_fedprox():
    fedprox, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedprox.yaml")
    # fedprox, log = _test_algo("./tests/configs/exp.yaml",
    #                           "./tests/configs/alg/fedprox.yaml", oncpu=False)


def test_fedper():
    fedper, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedper.yaml")
    # fedper, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedper.yaml", oncpu=False)


def test_fedrep():
    fedrep, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedrep.yaml")
    # fedrep, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedrep.yaml", oncpu=False)


def test_fedrod():
    fedrod, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedrod.yaml")
    # fedrod, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedrod.yaml", oncpu=False)


def test_fedrs():
    fedrs, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedrs.yaml")
    # fedrs, log = _test_algo("./tests/configs/exp.yaml",
    #                         "./tests/configs/alg/fedrs.yaml", oncpu=False)


def test_fedsam():
    fedsam, log = _test_algo(
        "./tests/configs/exp.yaml", "./tests/configs/alg/fedsam.yaml", tol=1e-3
    )
    # fedsam, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedsam.yaml", oncpu=False)


def test_fedsgd():
    fedsgd, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/fedsgd.yaml")
    # fedsgd, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/fedsgd.yaml", oncpu=False)


def test_gear():
    gear, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/gear.yaml")
    # gear, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/gear.yaml", oncpu=False)


def test_kafe():
    kafe, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/kafe.yaml")


def test_lgfedavg():
    lgfedavg, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/lg_fedavg.yaml")
    # lgfedavg, log = _test_algo("./tests/configs/exp.yaml",
    #                            "./tests/configs/alg/lg_fedavg.yaml", oncpu=False)


def test_moon():
    moon, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/moon.yaml")
    # moon, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/moon.yaml", oncpu=False)


def test_per_fedavg():
    per_fedavg, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/per_fedavg.yaml")
    # per_fedavg, log = _test_algo("./tests/configs/exp.yaml",
    #                              "./tests/configs/alg/per_fedavg.yaml", oncpu=False)
    per_fedavg, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/per_fedavg2.yaml")
    # per_fedavg, log = _test_algo("./tests/configs/exp.yaml",
    #                              "./tests/configs/alg/per_fedavg2.yaml", oncpu=False)


def test_pfedme():
    pfedme, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/pfedme.yaml")
    # pfedme, log = _test_algo("./tests/configs/exp.yaml",
    #                          "./tests/configs/alg/pfedme.yaml", oncpu=False)


def test_scaffold():
    scaffold, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/scaffold.yaml")
    # scaffold, log = _test_algo("./tests/configs/exp.yaml",
    #                            "./tests/configs/alg/scaffold.yaml", oncpu=False)


def test_superfed():
    superfed, log = _test_algo("./tests/configs/exp.yaml", "./tests/configs/alg/superfed.yaml", 3)
    # superfed, log = _test_algo("./tests/configs/exp.yaml",
    #                            "./tests/configs/alg/superfed.yaml", 3, oncpu=False)


if __name__ == "__main__":
    # test_centralized_fl()
    # test_apfl()
    # test_ccvr()
    # test_ditto()
    # test_fat()
    # test_fedamp()
    # test_fedavg()
    # test_fedavgm()
    # test_fedaws()
    # test_fedbabu()
    # test_fedbn()
    # test_feddyn()
    # test_fedexp()
    # test_fedhp()
    # test_fedlc()
    # test_fedld()
    # test_fednh()
    # test_fednova()
    # test_fedopt()
    # test_fedper()
    # test_fedprox()
    # test_fedproto()
    # test_fedrep()
    # test_fedrod()
    # test_fedrs()
    test_fedsam()
    # test_fedsgd()
    # test_gear()
    # test_kafe()
    # test_lgfedavg()
    # test_moon()
    # test_per_fedavg()
    # test_pfedme()
    # test_scaffold()
    # test_superfed()



"""Implementation of the FedExP [FedExP23]_ algorithm.

References:
    .. [FedExP23] Divyansh Jhunjhunwala, Shiqiang Wang, and Gauri Joshi.
       FedExP: Speeding Up Federated Averaging via Extrapolation. In ICLR (2023).
       URL: https://arxiv.org/abs/2301.09604
"""

import sys
from typing import Collection, Sequence

import torch
from torch.nn import Module

sys.path.append(".")
sys.path.append("..")

from ..algorithms import CentralizedFL  # NOQA
from ..client import Client  # NOQA
from ..server import Server  # NOQA
from ..utils.model import flatten_parameters  # NOQA

__all__ = ["FedExPServer", "FedExP"]


def _compute_eta(w: torch.Tensor, wi: list[torch.Tensor], eps: float = 1e-4) -> float:
    delta_bar = torch.mean(w - torch.stack(wi), dim=0)
    sum_norm_delta_i = torch.sum(torch.norm(w - torch.stack(wi)) ** 2, dim=0)
    norm_delta_bar = torch.norm(delta_bar) ** 2
    eta = torch.max(
        sum_norm_delta_i / (2 * len(wi) * (norm_delta_bar + eps)), torch.FloatTensor([1.0])
    )
    return eta.item()


class FedExPServer(Server):

    @torch.no_grad()
    def aggregate(self, eligible: Sequence[Client], client_models: Collection[Module]) -> None:
        w = flatten_parameters(self.model)
        client_models = list(client_models)
        wi = [flatten_parameters(client_model) for client_model in client_models]
        eta = _compute_eta(w, wi)

        self.hyper_params.update(lr=eta)
        super().aggregate(eligible, client_models)


class FedExP(CentralizedFL):

    def get_server_class(self) -> type[Server]:
        return FedExPServer
